/* sane - Scanner Access Now Easy.  Copyright (C) 2001 Henning Meier-Geinitz.
   HP Scanjet modifications Copyright (C) 2001 Stefan Nilsen

   Based on documentation from Mustek and National Semiconductor.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc., 59
   Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for additional
   uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files to
   produce an executable, this does not by itself cause the resulting
   executable to be covered by the GNU General Public License.  Your use of
   that executable is in no way restricted on account of linking the SANE
   library code into it.

   This exception does not, however, invalidate any other reasons why the
   executable file might be covered by the GNU General Public License.

   If you submit changes to SANE to the maintainers to be included in a
   subsequent release, you agree by submitting the changes that those changes
   may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice whether
   to permit this exception to apply to your modifications.  If you do not
   wish that, delete this exception notice.

   This file is originally a test program for the Mustek Bearpaw 1200.
   I (Stefan Nilsen) have now modified it for HP ScanJet 2200C.

   Last updated: 2001-07-11
*/
/* define TEST if you want this to be the standalone version */
#ifndef V_MAJOR 
#define TEST
#endif
#ifdef TEST
#define BACKEND_NAME LM9382
#define STUBS
#endif
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <error.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sane/sane.h>
#include <sane/sanei.h>
// #include <sane/sanei_debug.h>
#include <sane/saneopts.h>
#include <sane/sanei_config.h>
#include <sane/sanei_backend.h>
#include "lm9832.h"

#define MIN(a,b)	((a) < (b) ? (a) : (b))
#define MAX(a,b)	((a) > (b) ? (a) : (b))

#define HIBYTE(n) (((n) >> 8) & 0xff)
#define LOBYTE(n) ((n) & 0xff)

#define RETURN_IF_ERROR(func)         \
    {                                 \
      SANE_Status status;             \
      status = func;                  \
      if (status != SANE_STATUS_GOOD) \
      return status;                  \
    }

#define KLUDGE_EPSON_1250      
#define DBGLVL 3
#undef DBG
#define DBG(level, msg, args...) \
  do {if (level <= DBGLVL) fprintf (stderr, msg, ##args);} while (0)

#define SANEI_LM9832_MAX_REG 0x7f
#define MM_PER_INCH 25.4
static SANE_Word sanei_lm9832_max_transfer_size = 8096;

#define PIXEL_RATE_3_CHANNELS 0
#define LINE_RATE_3_CHANNELS 1

/* scanner hardware details */
/* convenience macros for scanner details */
#define DPI(d) (d),{8,d/12,d/8,d/6,d/4,d/3,d/2,3*d/2,d}
#define RANGE(d) {0,SANE_FIX(d),0}

#define SCANNER_MAKE_COUNT 2
const scanner_chars sane_scanner_details[] = {
  { /* Epson Perfection 1250 */
    "ep1250",
    "Epson",
    "Perfection 1250",
    "flatbed scanner",
    600, /* fspi*/
    DPI(1200), /* dpi */
    RANGE(216), /* x range */
    RANGE(300), /* y range */
    0.00666666666666667, /* sensor sep in inches */
    168, /* y offset of scan area */
    4}, /* tr_pulse_length */
  {/*HP Scanjet2200c*/
    "hp2200",
    "HP",
    "Scanjet 2200c",
    "flatbed scanner",
    300, /* fspi*/
    DPI(600), /* dpi */
    RANGE(216), /* x range (value is max in mm)*/
    RANGE(280), /* y range (value is max in mm)*/
    0.00666666666666667, /* sensor sep in inches */
    168, /* y offset of scan area */
    6} /* tr_pulse_length */
};

lm9832_Device* sane_lm9832_devices;
static int num_devices;
static char little_endian;
static char interleave_colour=0;
struct _measured_parameters_t
{
  unsigned int datalink_bandwidth;
  struct
  {
    int red;
    int green;
    int blue;
  } coarse_calibration_data;
  struct
  {
    int *pRedOffset;
    int *pGreenOffset;
    int *pBlueOffset;
    int *pRedGain;
    int *pGreenGain;
    int *pBlueGain;
  } fine_calibration_data;
  int max_integration_time;
  int color_mode;
};

struct coarse_t
{
  int min_red;
  int min_green;
  int min_blue;
  int max_red;
  int max_green;
  int max_blue;
  int red_gain;
  int red_offset;
  int green_gain;
  int green_offset;
  int blue_gain;
  int blue_offset;
};

typedef struct _measured_parameters_t measured_parameters_t;

static const double hdpi_mapping[8] = {1, 1.5, 2, 3, 4, 6, 8, 12};




SANE_Status
sanei_lm9832_open (lm9832_Scanner *s, SANE_String_Const devicename,
		   SANE_Word max_transfer_size)
{
  assert(s != 0);

  DBG(5, "sanei_lm9832_open: devicename=`%s', max_transfer_size=%d\n",
      devicename, max_transfer_size);
  sanei_lm9832_max_transfer_size = max_transfer_size;
  s->fd = open (devicename, O_RDWR | O_EXCL);
  if (s->fd < 0)
    {
      DBG(4, "sanei_lm9832_open: open failed (%s)\n", strerror(errno));
      if (errno == EACCES)
	return SANE_STATUS_ACCESS_DENIED;
      else if (errno == EBUSY)
	return SANE_STATUS_DEVICE_BUSY;
      else
	return SANE_STATUS_INVAL;
    }
  DBG(4, "sanei_lm9832_open: succeeded, fd=%d, devicename=`%s'\n", s->fd,
      devicename);
  return SANE_STATUS_GOOD;
}

SANE_Status
sanei_lm9832_close (lm9832_Scanner *s)
{
  SANE_Int status;

  assert(s != 0);

  DBG(5, "sanei_lm9832_close: fd=%d\n", s->fd);
  status = close (s->fd);
  if (status < 0)
    {
      DBG(4, "sanei_lm9832_close: close failed (%s)\n", strerror(errno));
      return SANE_STATUS_INVAL;
    }
  DBG(4, "sanei_lm9832_close: succeded\n");
  return SANE_STATUS_GOOD;
}


SANE_Status
sanei_lm9832_write (lm9832_Scanner *s, SANE_Byte reg, SANE_Byte *buffer, 
		    SANE_Word len, SANE_Bool increment)
{
  SANE_Byte command_buffer[sanei_lm9832_max_transfer_size + 4];
  SANE_Int result;
  SANE_Word bytes, max_len;
  
  assert(s != 0);

  DBG (5, "sanei_lm9832_write: fd=%d, reg=%d, len=%d, increment=%d\n", s->fd,
       reg, len, increment);

  for (bytes = 0; len > 0; )
    {
      max_len = MIN(len, sanei_lm9832_max_transfer_size);
      command_buffer[0] = 0;                    /* write */
      command_buffer[1] = reg;                  /* LM9832 register */
      if (increment)
	{
	  command_buffer[0] += 0x02;             /* increase reg? */
	  command_buffer[1] += bytes; 

	}
      command_buffer[2] = (max_len >> 8) & 0xff; /* bytes to write MSB */
      command_buffer[3] = max_len & 0xff;        /* bytes to write LSB */
      memcpy (command_buffer + 4, buffer + bytes, max_len);
      result = write (s->fd, command_buffer, len + 4);
      if (result < 0)
	{
	  DBG (1, "sanei_lm9832_write: write failed (%s)\n", strerror(errno));
	  return SANE_STATUS_IO_ERROR;
	}
      if (result != (max_len + 4))
	{
	  DBG (2, "sanei_lm9832_write: short write (%d/%d)\n", result,
	       max_len + 4);
	  if (result < 4)
	    {
	      DBG (2, "sanei_lm9832_write: couldn't even send command\n");
	      return SANE_STATUS_IO_ERROR;
	    }
	  DBG (2, "sanei_lm9832_write: trying again\n");
	}
      len -= (result - 4);
      bytes += (result - 4);
    }
  DBG(5, "sanei_lm9832_write: succeeded\n");
  return SANE_STATUS_GOOD;
}


SANE_Status
sanei_lm9832_write_byte (lm9832_Scanner *s, SANE_Byte reg, SANE_Byte value)
{
  assert(s != 0);
  return sanei_lm9832_write (s, reg, &value, 1, 0);
}

/*
#define sanei_lm9832_write_byte(fd, reg, value) \
   sanei_lm9832_write (fd, reg, &value, 1, 0)
*/
SANE_Status
sanei_lm9832_read (lm9832_Scanner *s, SANE_Byte reg, SANE_Byte *buffer, 
		   SANE_Word len, SANE_Bool increment)
{
  SANE_Byte command_buffer[4];
  SANE_Int result;
  SANE_Word bytes, max_len, read_bytes;

  assert(s != 0);

  DBG (5, "sanei_lm9832_read: fd=%d, reg=%d, len=%d, increment=%d\n", s->fd,
       reg, len, increment);

  if (!buffer)
    {
      DBG (1, "sanei_lm9832_read: error: buffer == NULL\n");
      return SANE_STATUS_NO_MEM;
    }

  for (bytes = 0; len > 0; )
    {
      max_len = MIN(len, sanei_lm9832_max_transfer_size);
      command_buffer[0] = 1;                /* read */
      command_buffer[1] = reg;               /* LM9832 register */
      if (increment)
	{
	  command_buffer [0] += 0x02;
	  command_buffer [1] += bytes;    
	}
      command_buffer[2] = (max_len >> 8) & 0xff; /* bytes to read MSB */
      command_buffer[3] = max_len & 0xff;        /* bytes to read LSB */
      DBG (5, "sanei_lm9832_read: writing command: "
	   "%02x %02x %02x %02x\n", command_buffer[0], command_buffer[1],
	   command_buffer[2], command_buffer[3]);
      result = write (s->fd, command_buffer, 4);
      if (result < 0)
	{
	  DBG (1, "sanei_lm9832_read: failed while writing command (%s)\n",
	       strerror(errno));
	  return SANE_STATUS_IO_ERROR;
	}
      if (result != 4)
	{
	  DBG (2, "sanei_lm9832_read: short write while writing command "
	       "(%d/4)\n", result);
	  return SANE_STATUS_IO_ERROR;
	}

      read_bytes = 0;
      do
	{
	  result = read (s->fd, buffer + bytes + read_bytes, 
			 max_len - read_bytes);
	  if (result < 0)
	    {
	      DBG (1, "sanei_lm9832_read: failed while reading (%s)\n",
		   strerror(errno));
	      return SANE_STATUS_IO_ERROR;
	    }
	  if (result == 0)
	    {
	      DBG (4, "sanei_lm9832_read: got eof\n");
	      return SANE_STATUS_EOF;
	    }
	  read_bytes += result;
	  DBG (5, "sanei_lm9832_read: read %d bytes\n", result);
	  
	  {
	    SANE_Int value_nr;
	    SANE_String value_text [100];
	    SANE_String value [4];

	    value_text[0] = '\0';
	    for (value_nr = 0; value_nr < result; value_nr++)
	      {
		sprintf ((SANE_String) value, "%02x ", 
			 buffer[value_nr + bytes]);
		strcat ((SANE_String) value_text, (SANE_String) value);
		if (((value_nr % 16) == 15) || (value_nr == (result - 1)))
		  {
		    DBG (5, "%02x: %s\n", (value_nr / 0x10) * 0x10,
			 (char *)value_text);
		    value_text[0] = '\0';
		  }
	      }
	  }

	  if (read_bytes != max_len)
	    {
	      DBG (2, "sanei_lm9832_read: short read (%d/%d)\n", result,
		   max_len);
	      usleep (10000);
	      DBG (2, "sanei_lm9832_read: trying again\n");
	    }
	}
      while (read_bytes < max_len);
      bytes += (max_len);
      len -= (max_len);
    }
  DBG(5, "sanei_lm9832_read: succeeded\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
sanei_lm9832_read_noblock (lm9832_Scanner *s, SANE_Byte reg, SANE_Byte *buffer, 
			   SANE_Word len, SANE_Bool increment, SANE_Word *read_count)
{
  SANE_Byte command_buffer[4];
  SANE_Int result;
  SANE_Word bytes, max_len, read_bytes;

  assert(s != 0);

  DBG (5, "sanei_lm9832_read_noblock: fd=%d, reg=%d, len=%d, increment=%d\n", s->fd,
       reg, len, increment);

  if (!buffer)
    {
      DBG (1, "sanei_lm9832_read_noblock: error: buffer == NULL\n");
      return SANE_STATUS_NO_MEM;
    }
  if (!read)
    {
      DBG (1, "sanei_lm9832_read_noblock: error: read == NULL\n");
      return SANE_STATUS_NO_MEM;
    }

  for (bytes = 0; len > 0; )
    {
      max_len = MIN(len, sanei_lm9832_max_transfer_size);
      command_buffer[0] = 1;                /* read */
      command_buffer[1] = reg;               /* LM9832 register */
      if (increment)
	{
	  command_buffer [0] += 0x02;
	  command_buffer [1] += bytes;    
	}
      command_buffer[2] = (max_len >> 8) & 0xff; /* bytes to read MSB */
      command_buffer[3] = max_len & 0xff;        /* bytes to read LSB */
      DBG (5, "sanei_lm9832_read_noblock: writing command: "
	   "%02x %02x %02x %02x\n", command_buffer[0], command_buffer[1],
	   command_buffer[2], command_buffer[3]);
      result = write (s->fd, command_buffer, 4);
      if (result < 0)
	{
	  DBG (1, "sanei_lm9832_read_noblock: failed while writing command (%s)\n",
	       strerror(errno));
	  return SANE_STATUS_IO_ERROR;
	}
      if (result != 4)
	{
	  DBG (2, "sanei_lm9832_read_noblock: short write while writing command "
	       "(%d/4)\n", result);
	  return SANE_STATUS_IO_ERROR;
	}

      read_bytes = 0;
      do
	{
	  result = read (s->fd, buffer + bytes + read_bytes, 
			 max_len - read_bytes);
	  if (result < 0)
	    {
	      DBG (1, "sanei_lm9832_read_noblock: failed while reading (%s)\n",
		   strerror(errno));
	      return SANE_STATUS_IO_ERROR;
	    }
	  if (result == 0)
	    {
	      DBG (4, "sanei_lm9832_read_noblock: got eof\n");
	      return SANE_STATUS_EOF;
	    }
	  read_bytes += result;
	  DBG (5, "sanei_lm9832_read_noblock: read %d bytes\n", result);
	  
	  {
	    SANE_Int value_nr;
	    SANE_String value_text [100];
	    SANE_String value [4];

	    value_text[0] = '\0';
	    for (value_nr = 0; value_nr < result; value_nr++)
	      {
		sprintf ((SANE_String) value, "%02x ", 
			 buffer[value_nr + bytes]);
		strcat ((SANE_String) value_text, (SANE_String) value);
		if (((value_nr % 16) == 15) || (value_nr == (result - 1)))
		  {
		    DBG (5, "%02x: %s\n", (value_nr / 0x10) * 0x10,
			 (char *)value_text);
		    value_text[0] = '\0';
		  }
	      }
	  }

	  if (read_bytes != max_len)
	    {
	      DBG (2, "sanei_lm9832_read_noblock: short read (%d/%d)\n", result,
		   max_len);
	      *read_count = bytes + read_bytes;
	      return SANE_STATUS_GOOD;
	    }
	}
      while (read_bytes < max_len);
      bytes += (max_len);
      len -= (max_len);
      *read_count = bytes;
    }
  DBG(5, "sanei_lm9832_read_noblock: succeeded\n");
  return SANE_STATUS_GOOD;
}


/*
SANE_Status
sanei_lm9832_read_byte (SANE_Int fd, SANE_Byte reg, SANE_Byte * value)
{
  return sanei_lm9832_read (fd, reg, value, 1, 0);
}
*/

#define sanei_lm9832_read_byte(s, reg, value) \
  sanei_lm9832_read (s, reg, value, 1, 0)



SANE_Status
sanei_lm9832_detect (lm9832_Scanner *s, SANE_Byte *version)
{
	SANE_Byte value;
	DBG(5, "sanei_lm9832_detect\n");
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0x00));
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x08, 0x04));
	RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x07, &value));
	RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x08, &value));
	RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x69, &value));
	value &= 7;
	if (version)
		*version = value;
	if (4 == value) {
		DBG(4, "sanei_lm9832_detect: found LM9832 (v%d)\n", value);
		return SANE_STATUS_GOOD;
	} else if (3 == value) {
		DBG(4, "sanei_lm9832_detect: found LM9832 (v%d)\n", value);
		return SANE_STATUS_GOOD;
	} else if (2 == value) {
		DBG(4, "sanei_lm9832_detect: found unsupported LM9830 (v%d)\n", value);
		return SANE_STATUS_INVAL;
	} else {
		DBG(4, "sanei_lm9832_detect: Unknown chip (v%d)\n", value);
		return SANE_STATUS_INVAL;
	}
}

SANE_Status
sanei_lm9832_reset (lm9832_Scanner *s)
{
  SANE_Byte value;

  assert(s != 0);

  DBG(5, "sanei_lm9832_reset\n");
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0));
  RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x18, &value));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x18, 0x18));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07,0x20));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x18, value));
  RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x07, &value));
  if (value != 0)
    {
      DBG(1, "sanei_lm9832_reset: reset wasn't successful, status=%d\n",
	  value);
      return SANE_STATUS_INVAL;
    }
  DBG(4, "sanei_lm9832_reset: succeeded\n");
  
  return SANE_STATUS_GOOD;
}

/***************************************************************************/

typedef enum
{
  lineart,
  gray1,
  gray2,
  gray4,
  gray8,
  color8,
  color14
} 
LM9832_Scan_Mode;



static unsigned char
getreg (lm9832_Scanner *s, unsigned char reg)
{
  unsigned char reg_value;

  assert(s != 0);

  sanei_lm9832_read (s, reg, &reg_value,
		     1, SANE_TRUE);
  return reg_value;
}

static void
setreg (lm9832_Scanner *s, unsigned char reg, unsigned char reg_value)
{
  assert(s != 0);

  sanei_lm9832_write_byte (s, reg, reg_value);
}

/* --- Scanner Dependent Hardware config --- */

/* Misc IO settings */
static SANE_Status
misc_io (lm9832_Scanner *s) {
  /*	assert(s != 0);*/

  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x59, 0x41));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x5a, 0x44));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x5b, 0x49));
    break;
  case LM9832_HP2200:
  default:
    /* HP Scanjet 2200C */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x59, 0xcc));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x5a, 0xbc));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x5b, 0xbb));
    break;
  }
  return SANE_STATUS_GOOD;
}

/* Sensor settings (CCD) */
static SANE_Status
sensor_settings (lm9832_Scanner *s) {
  assert(s != 0);
  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0b, 0x02));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0c, 0x04));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0d, 0x7f));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0e, 0x13));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0f, 0x06));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x10, 0x16));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x11, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x12, 0x05));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x13, 0x0c));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x14, 0x17));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x15, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x16, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x17, 0x08));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x18, 0x14));
    break;
  case LM9832_HP2200:
    /* HP Scanjet 2200C Settings */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0b, 0x02));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0c, 0x7c));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0d, 0x3f));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0e, 0x15));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0f, 0x04));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x10, 0x16));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x11, 0x01));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x12, 0x02));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x13, 0x05));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x14, 0x06));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x15, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x16, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x17, 0x0a));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x18, 0x16));
    break;
  }
  return SANE_STATUS_GOOD;
}

/* Sensor Pixel Config */
static SANE_Status
sensor_pixel_settings (lm9832_Scanner *s) {
  assert(s != 0);
  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    /* Sensor pixel optical black pixels start */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1c, 0x00));
    /* Sensor pixel optical black pixels end */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1d, 0x42));
    /* Sensor pixel active pixels start MSB (0)*/
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1e, 0x00));
    /* Sensor pixel active pixels start LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1f, 0x45));
    /* Sensor pixel line end MSB (5369) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x20, 0x2a));
    /* Sensor pixel line end LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x21, 0x0e));
    break;
  case LM9832_HP2200:
    /* HP Scanjet 2200C */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1c, 0x14));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1d, 0xf7));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1e, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1f, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x20, 0x13));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x21, 0xf0));
    break;
  }
  return SANE_STATUS_GOOD;
}

/* Stepper phase correction */
static SANE_Status
stepper_phase_settings (lm9832_Scanner *s) {
  assert(s != 0);
  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    /* Stepper phasse correction */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1a, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1b, 0x01));
    break;
  case LM9832_HP2200:
    /* HP Scanjet 2200C */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1a, 0x01));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1b, 0x01));
    break;
  }
  return SANE_STATUS_GOOD;
}
		
/* Motor settings */
static SANE_Status
motor_settings (lm9832_Scanner *s) {
  assert(s != 0);
  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    /* Stepper motor n line skipping (0) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x43, 0x00));
    /* Stepper motor m line skipping (0) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x44, 0x00));
    /* Microstepping */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x45, 0x03));
    /* Scanning step size MSB (400) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x46, 0x03));
    /* Scanning step size LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x47, 0x59));
    /* Fast feed Step size MSB (400) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x48, 0x00));
    /* Fast feed Step size LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x49, 0xff));
    /* Full steps to skip at start MSB (0) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4a, 0x00));
    /* Full steps to skip at start LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4b, 0x00));
    /* Full steps after paper sense 2 MSB (0) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4c, 0x00));
    /* Full steps after paper sense 2 LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4d, 0x00));
    /* Pause scanning (0x7f) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4e, 0x8e));
    /* Resume scanning (0x01) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4f, 0x48));
    
    /* Stepper motor full steps to reverse */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x50, 0x00));
    /* Stepper motor acceleration profile (3,3,3) + phase diff MSB (0) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x51, 0x0c));
    /* Stepper motor phase diff MedSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x52, 0x2a));
    /* Stepper motor phase diff LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x53, 0x20));
    /* Stepper motor lines to process, line skipping phase + delay */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x54, 0x00));
    /* Stepper motor kickstart steps (3), hold current timeout (2) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x55, 0x0f));
    /* Stepper motor PWM frequency */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x56, 0x02));
    /* Stepper motor PWM set duty cycle (32) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x57, 0x01));
    break;
  case LM9832_HP2200:
    /* HP Scanjet 2200C */
    
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x43, 0x1b));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x44, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x45, 0x03));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x46, 0x01));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x47, 0x8c));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x48, 0x01));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x49, 0x8c));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4a, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4b, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4c, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4d, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4e, 0x0b));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4f, 0x01));
    
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x50, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x51, 0xfc));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x52, 0x10));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x53, 0x79));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x54, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x55, 0x10));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x56, 0x08));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x57, 0x3c));
    break;
  }
  return SANE_STATUS_GOOD;
}


/* Sensor settings (Paper) */
static SANE_Status
paper_sensor (lm9832_Scanner *s) {
  assert(s != 0);
  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    /* Epson Perfection 1250*/ 
    /* uses level detection on PAPER SENSE 1*/
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x58, 0x09));
    break;
  case LM9832_HP2200:
    /* HP Scanjet 2200C */
    /* HP Uses 0x17 here (edge triggered), but I find that insecure.
       I use level sensitive to actually se if it is home now */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x58, 0x15));
    break;
  }
  return SANE_STATUS_GOOD;
}

/* Color Mode settings */
static SANE_Status
color_mode (lm9832_Scanner *s) {
  assert(s != 0);
  
  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    /* Color mode (AFE=0, gray=8 (green), TRred=0, TRgreen=0, 
       TRblue=0) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x26, 0x00));
    /* Color mode line rate TR drop (red=0, green=0, blue=0, tto=0) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x27, 0x40));
    break;
  case LM9832_HP2200:
    /* HP Scanjet 2200C */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x26, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x27, 0x00));
    break;
  }
  return SANE_STATUS_GOOD;
}

/* Illumination settings (LAMP) */
static SANE_Status
illumination_mode (lm9832_Scanner *s) {
  assert(s != 0);
  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    /* Illumination mode (mode 1) */
    /* Hmm, Usb sniffing thinks 3 is a great way to go here. 
       I'm not convinced. Lets run with 1, with everything on 
       all the time */
    /* Actually, I now think that the Lamp is managed by Misc6. Doh!
     * so what are all the below doing?? don't seem to make any difference... */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x29, 0x03));
    /* Illumination lamp(g) PWM MSB (4095) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2A, 0x00));
    /* Illumination lamp(g) PWM LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2B, 0x00));
    /* Illumination lamp(r) on MSB (1) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2c, 0x3f));
    /* Illumination lamp(r) on LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2d, 0xff));
    /* Illumination lamp(r) off MSB (1) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2e, 0x00));
    /* Illumination lamp(r) off LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2f, 0x00));
    /* Illumination lamp(g) on MSB (1) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x30, 0x3f));
    /* Illumination lamp(g) on LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x31, 0xff));
    /* Illumination lamp(g) off MSB (1) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x32, 0x00));
    /* Illumination lamp(g) off LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x33, 0x00));
    /* Illumination lamp(b) on MSB (1) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x34, 0x3f));
    /* Illumination lamp(b) on LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x35, 0xff));
    /* Illumination lamp(b) off MSB (1) */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x36, 0x00));
    /* Illumination lamp(b) off LSB */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x37, 0x00));
    break;
  case LM9832_HP2200:
    /* HP Scanjet 2200C */
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x29, 0x01));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2A, 0x00)); 
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2B, 0x00)); 
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2c, 0x00)); 
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2d, 0x01)); 
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2e, 0x3f));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x2f, 0xff));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x30, 0x3f));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x31, 0xff));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x32, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x33, 0x01));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x34, 0x3f));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x35, 0xff));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x36, 0x00));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x37, 0x01));
    break;
  }
  return SANE_STATUS_GOOD;
}

/* ---------------------------------- */

static SANE_Status
led (lm9832_Scanner *s, SANE_Bool on)
{
  SANE_Byte value;
  
  assert(s != 0);
  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x59, &value));
    /* perfection uses misc 6?*/
    if (on)
      value |= (1 << 3);
    else
      value &= ~(1 << 3);
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x59, value));
    break;
  case LM9832_HP2200:
    RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x5b, &value));
    if (on) 
      value |= (1 << 3);
    else
      value &= ~(1 << 3);
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x5b, value));
    break;
  }
  DBG(5, "led: LED is %s\n", on ? "on" : "off");
  return SANE_STATUS_GOOD;
}

static SANE_Status
wait_scanner_ready (lm9832_Scanner *s)
{
  SANE_Byte value;
  SANE_Word delay;
  
  assert(s != 0);
  
  DBG(5, "wait_scanner_ready\n");
  for (delay = 0; delay < 300; delay++)
    {
      RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x07, &value));
      if (value == 0)
	{
	  DBG(4, "wait_scanner_ready: scanner is ready (%d tries)\n", delay);
	  return SANE_STATUS_GOOD;
	}
      if (value >= 8)
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0x00));
      usleep (100000);
    }
  DBG(1, "wait_scanner_ready: scanner is not ready (%d tries)\n", delay);
  return SANE_STATUS_IO_ERROR;
}

static SANE_Status
check_sensor_home (lm9832_Scanner *s, SANE_Bool *home)
{
  SANE_Byte value;
	
  assert(s != 0);
	
  RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x02, &value));
  if ((value & 0x01) == 0x01)
    *home = SANE_TRUE;
  else
    *home = SANE_FALSE;
  DBG(4, "check_sensor_home: sensor was %sat home\n", 
      *home ? "" : "not ");
  return SANE_STATUS_GOOD;
}

static SANE_Status
paper_sensor1_on(lm9832_Scanner *s) {
  SANE_Byte value;
  
  assert(s != 0);
  
  /* Turn on paper sense 1 */
  RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x58, &value));
  value |= 4;
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x58, value));
  
  return SANE_STATUS_GOOD;
}

static SANE_Status
paper_sensor1_off(lm9832_Scanner *s) {
  SANE_Byte value;
  
  assert(s != 0);
  
  /* Turn off paper sense 1 */
  RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x58, &value));
  value &= (0xff ^ 0x04);
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x58, value));
  
  return SANE_STATUS_GOOD;
}

static SANE_Status
home_sensor (lm9832_Scanner *s)
{
  SANE_Byte value;
  SANE_Bool home;
  
  assert(s != 0);
  
  /* Turn on paper sense 1 */
  RETURN_IF_ERROR(paper_sensor1_on( s));
  /* THIS IS IMPORTANT. CAN DESTROY SCANNER WITHOUT THIS CHECK */
  RETURN_IF_ERROR(check_sensor_home (s, &home));
  if (SANE_TRUE == home) {
    DBG(4, "sensor is already home\n");
    return SANE_STATUS_GOOD;
  } else {
    DBG(4, "sensor is not home. will continue with homing.\n");
  }
  
  
  DBG(5, "home_sensor\n");
  RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x07, &value));
  if (value & 0x02)
    {
      DBG(5, "home_sensor: sensor already homing\n");
      /* Wait for scanner to return */
      RETURN_IF_ERROR(wait_scanner_ready(s));
      return SANE_STATUS_GOOD;
    }
  /* home sensor */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0x02));
  DBG(4, "home_sensor: sensor homing\n");
  /* Wait for scanner to return */
  RETURN_IF_ERROR(wait_scanner_ready(s));
  return SANE_STATUS_GOOD;
}

/* Fast Forward n lines */
static SANE_Status
move_lines_ff (lm9832_Scanner *s, SANE_Word n) {
  SANE_Word n_real;
  SANE_Byte value;

  assert(s != 0);
  
  n_real = n;
  if (n_real > 6000) {
    n_real = 6000;
  }
  DBG(5, "fast forward %d lines\n", n);
  
  RETURN_IF_ERROR(wait_scanner_ready (s));
  /* Move forward scan slider for n lines as given in 4A-4B */
  value = 0x7f & HIBYTE(n_real);
  DBG(1, "MSB %d lines\n", value);
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4A, value));
  value = LOBYTE(n_real);
  DBG(1, "LSB %d lines\n", value);
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4B, value));
  /* 0x05 = forward, 0x06 = reverse */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0x05));
  usleep (5 * 1000);
  
  RETURN_IF_ERROR(wait_scanner_ready (s));
  
  /* TODO: Restore old value */
  value = 0;
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4A, value));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4B, value));
  
  return SANE_STATUS_GOOD;
}

/* Fast Reverse n lines */
/* This can be dangerous if you move more backward then forward i guess. */
static SANE_Status
move_lines_fr (lm9832_Scanner *s, SANE_Word n) {
  SANE_Word n_real;
  SANE_Byte value;
  
  assert(s != 0);
  
  /* This function is only for test. Use max 32767 lines */
  n_real = n;
  if (n_real > 32767) {
    n_real = 32767;
  }
  DBG(5, "fast reverse %d lines\n", n);
  
  /* Move forward scan slider for n lines as given in 4A-4B */
  value = 0x7f & (n_real >> 8);
  DBG(1, "MSB %d lines\n", value);
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4A, value));
  value = 0xff & n_real;
  DBG(1, "LSB %d lines\n", value);
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4B, value));
  /* 0x05 = forward, 0x06 = reverse */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0x06)); 
  usleep (5 * 1000);
  
  RETURN_IF_ERROR(wait_scanner_ready (s));
  
  return SANE_STATUS_GOOD;
}

/* Fast Forward n secs */
static SANE_Status
move_secs_ff (lm9832_Scanner *s, SANE_Word n) {
  SANE_Word n_real; 
  SANE_Byte value;
  
  assert(s != 0);
  
  /* This function is only for test. Use max 5 secs */
  n_real = n;
  if (n_real > 5) {
    n_real = 5;
  }
  DBG(5, "fast forward %d secs\n", n_real);
  
  /* Turn off paper sense 1 */
  RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x58, &value));
  value &= ~4;
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x58, value));
  /* Move forward scan slider for 2 seconds */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0x01));
  usleep (n_real * 1000 * 1000);
  /* Turn on paper sense 1 */
  RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x58, &value));
  value |= 4;
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x58, value));
  /* Stop moving */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0x00));
  usleep (5 * 1000);
  
  RETURN_IF_ERROR(wait_scanner_ready (s));
  
  return SANE_STATUS_GOOD;
}

/* Lamp off  */
static SANE_Status
lamp_off (lm9832_Scanner *s) {
  DBG(5, "turn off lamp\n");
  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x5b, 0x14));
    break;
  case LM9832_HP2200:
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x29, 0x00));
    break;
  }
  return SANE_STATUS_GOOD;
}

/* Lamp on */
static SANE_Status
lamp_on (lm9832_Scanner *s) {
  DBG(5, "turn on lamp\n");
  /* Illumination mode (mode 1) */
  switch (s->hw->scanner_make) {
  case LM9832_EPSON1250:
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x5b, 0x94));
    break;
  case LM9832_HP2200:
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x29, 0x01));
    break;
  }
  return SANE_STATUS_GOOD;
}

static SANE_Status
print_registers (lm9832_Scanner *s)
{
  SANE_Int reg;
  SANE_String registers_text [100];
  SANE_String register_value [4];
  SANE_Byte register_byte [SANEI_LM9832_MAX_REG + 1];

  DBG(3, "print_registers: registers output:\n");
  registers_text[0] = '\0';
  /* reg 0 usually results in eof */
  sprintf ((SANE_String) register_value, "-- ");
  strcat ((SANE_String) registers_text, (SANE_String) register_value);
  RETURN_IF_ERROR(sanei_lm9832_read (s, 1, register_byte, 
				     SANEI_LM9832_MAX_REG, SANE_TRUE));
  
  for (reg = 1; reg <= SANEI_LM9832_MAX_REG; reg++)
    {
      sprintf ((SANE_String) register_value, "%02x ", register_byte[reg - 1]);
      strcat ((SANE_String) registers_text, (SANE_String) register_value);
      if (((reg % 16) == 15) || (reg == SANEI_LM9832_MAX_REG))
	{
	  DBG (3, "%02x: %s\n", (reg / 0x10) * 0x10, (char *)registers_text);
	  registers_text[0] = '\0';
	}

    }
  DBG(5, "print_registers: succeeded\n");
  return SANE_STATUS_GOOD;
}

static SANE_Status
motor (lm9832_Scanner *s, SANE_Bool on)
{
  if (on)
    {
      /* micro=1, current=1, phasea=0, phaseb=0, stat=1 */
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x45, 0x13));
      DBG(5, "motor: motor is on\n");
    }
  else
    {
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x45, 0x03));
      DBG(5, "motor: motor is off\n");
    }
  return SANE_STATUS_GOOD;
}

static SANE_Status
power (lm9832_Scanner *s, SANE_Bool on)
{
  SANE_Byte reg, value;
  
  assert(s != 0);
  
  DBG(4, "power: turn %s\n", on ? "on" : "off");
  
  if (on)
    {
      /* Reset */
      RETURN_IF_ERROR(sanei_lm9832_reset (s));
      /* ---------- Init asic ---------- */
      /* Reserved registers */
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0a, 0x00));
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x28, 0x00));
      /* Disable DRAM test */
      RETURN_IF_ERROR(sanei_lm9832_read_byte (s, 0x03, &value));
      value = value & ~ (1 << 6);
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x03, value));
      /* Clear data port registers */
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x04, 0x00));
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x05, 0x00));
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x06, 0x00));
      RETURN_IF_ERROR(sensor_settings (s));
      RETURN_IF_ERROR(sensor_pixel_settings (s));
      RETURN_IF_ERROR(color_mode (s));
      RETURN_IF_ERROR(illumination_mode (s));
      RETURN_IF_ERROR(stepper_phase_settings (s));
      RETURN_IF_ERROR(motor_settings (s));
      RETURN_IF_ERROR(paper_sensor (s));
      RETURN_IF_ERROR(misc_io (s));
      /* Test mode ADC Output Code MSB (0) */
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x5c, 0x00));
      /* Test mode ADC Output Code LSB */
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x5d, 0x00));
      /* Test mode (0) */
      RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x5e, 0x00));
      /* turn on LED */
      RETURN_IF_ERROR(led (s, SANE_TRUE));
      switch (s->hw->scanner_make) {
      case LM9832_EPSON1250:
	/* Master clock divider (2) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x08, 0x06));
	/* Horizontal res + datamode (horiz dpi div=0x05, pp=0x18, dm=0) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x09, 0x1d));
	/* Integration time adjust (0) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x19, 0x00));
	/* Pixel data range start MSB (0) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x22, 0x00));
	/* Pixel data range start LSB */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x23, 0x00));
	/* Pixel data range end MSB (5369-20) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x24, 0x14));
	/* Pixel data range end LSB */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x25, 0xe5));

	/* Default phase diff MedSB (70) */
	//		RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x52, 0x00));
	/* Default phase diff LSB () */
	//		RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x53, 0x46));
	/* Write reserved registers */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x28, 0x00));
	for (reg = 0x5f; reg < 0x80; reg++) {
	  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, reg, 0x00));
	}
	/* ---------- Init asic finished ---------- */
	/* RGB static offset */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x38, 0x00));
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x39, 0x00));
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3a, 0x00));
	/* RGB static gain */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3b, 0x30));
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3c, 0x30));
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3d, 0x30));
	/* Fixed offset coefficient MSB (0) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3e, 0x00));
	/* Fixed offset coefficient LSB */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3f, 0x00));
	/* Fixed multiplier coefficient MSB (16384) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x40, 0x40));
	/* Fixed multiplier coefficient LSB */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x41, 0x00));
	/* Digital data rate offset (0,0,0,0) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x42, 0x20));
	break;
      case LM9832_HP2200:
	/* Master clock divider (2) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x08, 0x02));
	/* Horizontal res + datamode (horiz dpi div=0x05, pp=0x18, dm=0) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x09, 0x1d));
	/* Integration time adjust (0) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x19, 0x00));
	/* Pixel data range start MSB (0) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x22, 0x01));
	/* Pixel data range start LSB */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x23, 0x35));
	/* Pixel data range end MSB (5369-20) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x24, 0x28));
	/* Pixel data range end LSB */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x25, 0xf5));

	/* Default phase diff MedSB (70) */
	//		RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x52, 0x00));
	/* Default phase diff LSB () */
	//		RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x53, 0x46));
	/* Write reserved registers */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x28, 0x00));
	for (reg = 0x5f; reg < 0x80; reg++) {
	  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, reg, 0x00));
	}
	/* ---------- Init asic finished ---------- */
	/* RGB static offset */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x38, 0x00));
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x39, 0x00));
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3a, 0x00));
	/* RGB static gain */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3b, 0x01));
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3c, 0x01));
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3d, 0x01));
	/* Fixed offset coefficient MSB (0) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3e, 0x00));
	/* Fixed offset coefficient LSB */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x3f, 0x00));
	/* Fixed multiplier coefficient MSB (16384) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x40, 0x40));
	/* Fixed multiplier coefficient LSB */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x41, 0x00));
	/* Digital data rate offset (0,0,0,0) */
	RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x42, 0x21));
	break;
      }
      /* Enable motor */
      RETURN_IF_ERROR(motor (s, SANE_TRUE));
      RETURN_IF_ERROR(lamp_on (s));
      RETURN_IF_ERROR(wait_scanner_ready (s));

      /* Home sensor */
      RETURN_IF_ERROR(home_sensor (s));
      RETURN_IF_ERROR(wait_scanner_ready (s));
    }
  else /* off */
    {
      /* shut down lamp */
      RETURN_IF_ERROR(lamp_off ( s));
      /* Disable motor */
      RETURN_IF_ERROR(motor (s, SANE_FALSE));
    }
  DBG(5, "power: succeded\n");
  return SANE_STATUS_GOOD;
}

static SANE_Status
wait_for_pause (lm9832_Scanner *s) {
  SANE_Byte register_03;
  int i;
  unsigned char scankb1;
  unsigned char scankb2;

  for (i=0; i<100; ++i) {
    scankb1 = getreg (s, 0x01);
    scankb2 = getreg (s, 0x01);
    register_03 = getreg (s, 0x03);
    if (scankb1 == scankb2) {
      if (register_03 & 0x10) {
	break;
      }
    }
    usleep(100L*1000L);
  }
  if (register_03 & 0x10) {
    return SANE_STATUS_GOOD;
  } else {
    return SANE_STATUS_INVAL;
  }
}

static SANE_Status
read_available_data (lm9832_Scanner *s, char *buffer, int *byte_count, int max_len)
{
  unsigned char scankb1;
  //  unsigned char scankb2;
  int to_read;
  int really_read;
  int chunk;

  assert (buffer != NULL);

  *byte_count = 0;
  //do {
  scankb1 = getreg (s, 0x01);
  //    scankb2 = getreg (s, 0x01);
  //    if (s->aborted_by_user)
  //      return SANE_STATUS_CANCELLED;
  // // usleep(125 * 1000);
  //  } while ((scankb1 != scankb2) /*|| (scankb1 < 2)*/ );
  DBG(4, "get_image_data: scankb1=%i\n", scankb1);

  to_read = MIN(scankb1 * 2 * 1024, max_len);
  DBG(4, "get_image_data: to_read=%i\n", to_read);

  while (to_read) {
    if (s->aborted_by_user) {
      return SANE_STATUS_CANCELLED;
    }
    chunk = (to_read > 0xffff) ? 0xffff : to_read;
    DBG(4, "get_image_data: chunk=%i\n", chunk);
    sanei_lm9832_read_noblock ( s, 0, buffer, 
				chunk, 0, &really_read);
    //		really_read = read (s->fd, buffer, chunk);
    if (really_read < 0) {
      perror ("read");
      return -1;
    }
    if (really_read > to_read) {
      perror ("USB stack read more bytes than requested!");
      return -1;
    }
    *byte_count += really_read;
    buffer += really_read;
    to_read -= really_read;
#ifdef DEBUG
    fprintf (stderr, "he leido %d bytes\n", really_read);
#endif
  }
  return SANE_STATUS_GOOD;
}

static SANE_Status
read_required_data (lm9832_Scanner *s, char *buffer, int required)
{
  unsigned char scankb1;
  unsigned char scankb2;
  int to_read;
  int read_count = 0;
  int really_read;
  int chunk;

  assert (buffer != NULL);

  while (required) {
    do {
      scankb1 = getreg (s, 0x01);
      scankb2 = getreg (s, 0x01);
      if (s->aborted_by_user)
	return SANE_STATUS_CANCELLED;
      // usleep(125 * 1000);
    } while ((scankb1 != scankb2) /*|| (scankb1 < 2)*/ );
    DBG(4, "get_image_data: scankb1=%i\n", scankb1);

    to_read = MIN(scankb1 * 2 * 1024, required);
    DBG(4, "get_image_data: to_read=%i\n", to_read);

    while (to_read) {
      if (s->aborted_by_user) {
	return SANE_STATUS_CANCELLED;
      }
      chunk = (to_read > 0xffff) ? 0xffff : to_read;
      DBG(4, "get_image_data: chunk=%i\n", chunk);
      sanei_lm9832_read_noblock ( s, 0, buffer, 
				  chunk, 0, &really_read);
      //		really_read = read (s->fd, buffer, chunk);
      if (really_read < 0) {
	perror ("read");
	return -1;
      }
      if (really_read > to_read) {
	perror ("USB stack read more bytes than requested!");
	return -1;
      }
      read_count += really_read;
      buffer += really_read;
      to_read -= really_read;
      required -= really_read;
#ifdef DEBUG
      fprintf (stderr, "he leido %d bytes\n", really_read);
#endif
    }
  }
  return SANE_STATUS_GOOD;
}

static SANE_Status
set_parameters (lm9832_Scanner *s,SANE_Word mode, SANE_Word dpi, 
		SANE_Word start_line, SANE_Word start_col, SANE_Word end_line,
		SANE_Word end_col, SANE_Bool preview) {
  SANE_Byte register_08;
  SANE_Byte register_09;
  SANE_Byte register_0A;
  SANE_Byte register_4A;
  SANE_Byte register_4B;
  SANE_Byte register_4C;
  SANE_Byte register_4D;
  SANE_Byte register_byte;
  const unsigned int available_memory = 296; /* KB, 1832 or 296 */
  unsigned long line_length;	/* KB */
  unsigned long pause_threshold; /* KB */
  unsigned long bytes_per_line; /* bytes */
  unsigned long data_pixels_start; /* Reg 22,23 */
  unsigned long data_pixels_end; /* Reg 24,25 */
  unsigned long data_pixels; /* data_pixels_end - data_pixels_start */
  unsigned int  channels; /* 1 for gray, 3 for all others */
  unsigned int  pixel_bits; /* 1,2,4,8,16(14bits) */
  SANE_Word  step_size; /* step size for this resolution */
  SANE_Word line_end;

  assert(s != 0);

  DBG(5, "set_parameters: fd=%d, mode=%d, dpi=%d, start_line=%d, "
      "start_col=%d, end_line=%d, end_col=%d, preview=%d\n", 
      s->fd, mode, dpi, start_line, start_col, end_line, end_col, preview);

  register_08 = 0;
  register_09 = 0;
  register_0A = 0;
  register_4A = 0;
  register_4B = 0;

  s->target_value_min.red = (50) >> 2;
  s->target_value_min.green = (50) >> 2;
  s->target_value_min.blue = (50) >> 2;
  s->target_value_low.red = (2048) >> 2;
  s->target_value_low.green = (2048) >> 2;
  s->target_value_low.blue = (2048) >> 2;
  s->target_value_high.red = (65535-16000) >> 2;
  s->target_value_high.green = (65535-16000) >> 2;
  s->target_value_high.blue = (65535-16000) >> 2;
  s->target_value_max.red = (65535-1000) >> 2;
  s->target_value_max.green = (65535-1000) >> 2;
  s->target_value_max.blue = (65535-1000) >> 2;

  s->scan_mode = mode;
  pixel_bits = mode;
  switch(mode) {
  case 8:
    register_09 |= 0x18; /* 8 bit color */
    channels = 3;
    break;
  case 16:
    register_09 |= 0x38; /* two byte data, 14 bits, (logical 16bits) */
    channels = 3;
    break;
  case 4:
    register_09 |= 0x10; /* 8 bit color */
    channels = 3;
    break;
  case 2:
    register_09 |= 0x08; /* 8 bit color */
    channels = 3;
    break;
  case 1:
    register_09 |= 0x00; /* 8 bit color */
    channels = 3;
    break;
  default:
    return SANE_STATUS_UNSUPPORTED;
  }

  switch(dpi*600/(s->details->dpi)) {
    /*FIXME - record changes to MCLK DIV - this is needed for fullstep checks...*/
  case 600:
    register_09 |= 0x00; 
    register_08 = 10;
    s->hdpi_dividerx10 = 10;
    break;
  case 400:
    register_09 |= 0x01;
    register_08 = 6;
    s->hdpi_dividerx10 = 15;
    break;
  case 300:
    register_09 |= 0x02;
    register_08 = 8;
    s->hdpi_dividerx10 = 20;
    break;
  case 200:
    register_09 |= 0x03;
    register_08 = 6;
    s->hdpi_dividerx10 = 30;
    break;
  case 150:
    register_09 |= 0x04;
    register_08 = 6;
    s->hdpi_dividerx10 = 40;
    break;
  case 100:
    register_09 |= 0x05;
    register_08 = 6;
    s->hdpi_dividerx10 = 60;
    break;
  case 75:
    register_09 |= 0x06;
    register_08 = 6;
    s->hdpi_dividerx10 = 80;
    break;
  case 50:
    register_09 |= 0x07;
    register_08 = 6;
    s->hdpi_dividerx10 = 120;
    break;
  default:
    return SANE_STATUS_UNSUPPORTED;
  }

  switch(preview) {
  case SANE_TRUE:
    // FIXME    register_0A |= 0x01;
    break;
  default:
    register_0A |= 0x00;
  }

  register_4A = HIBYTE(start_line);
  register_4B = LOBYTE(start_line);
  register_4C = HIBYTE(end_line-start_line);
  register_4D = LOBYTE(end_line-start_line);

  RETURN_IF_ERROR(wait_scanner_ready (s));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x08, register_08));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x09, register_09));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x0A, register_0A));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4A, register_4A));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4B, register_4B));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4C, register_4C));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4D, register_4D));

  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1c, 0x0e));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1d, 0x3f));

  //  line_end = MAX(end_col+20,((s->details->fspi * 4 * 256)/dpi)); /* make sure step size is at least 256 */
  line_end = 0x2a0e;
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1e, HIBYTE(start_col)));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x1f, LOBYTE(start_col)));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x20, HIBYTE(line_end)));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x21, LOBYTE(line_end)));

  /* Pixel data range start MSB (0) */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x22, HIBYTE(start_col)));
  /* Pixel data range start LSB */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x23, LOBYTE(start_col)));
  /* Pixel data range end MSB (5369-20) */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x24, HIBYTE(end_col)));
  /* Pixel data range end LSB */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x25, LOBYTE(end_col)));

  /* Set start/go thresholds for motor in line buffer */
  data_pixels_start = start_col;
  data_pixels_end = end_col;
  data_pixels = data_pixels_end - data_pixels_start;
  bytes_per_line = 2 * ( ( (10*data_pixels / s->hdpi_dividerx10) * channels * pixel_bits ) / 16 );
  line_length = bytes_per_line / 1024;
  pause_threshold = available_memory - (line_length+1);
  if (available_memory == 1832) {
    register_byte = pause_threshold/8;
  } else {
    register_byte = pause_threshold/2;
  }
  step_size = ((line_end + s->details->tr_pulse_length) * dpi * 1.0f) / (s->details->fspi * 4);
  DBG(3,"steps.. %d\n",step_size);
  if (step_size < 0x40) step_size = 0x0040;
  /* standard step size MSB*/
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x46, HIBYTE(step_size)));
  /* standard step size LSB*/
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x47, LOBYTE(step_size)));
  RETURN_IF_ERROR(wait_scanner_ready (s));
  register_byte = MIN(0x40, register_byte); /* There is some error in the formula above */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4e, 0x8e)); // fettled PU
  DBG(4, "set_parameters: threshold1 %i\n", (int)register_byte);
  register_byte = 2;
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x4f, register_byte));
  DBG(4, "set_parameters: threshold2 %i\n", (int)register_byte);

	/* Acceleration profile */	
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x50, 0x20));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x51, 0x0c));

	/* Save some parameters to paramblock */
  s->bytes_per_line = bytes_per_line;
  DBG(4, "set_parameters: bytes_per_line %li\n", bytes_per_line);

  DBG(5, "set_parameters: succeded\n");
  return SANE_STATUS_GOOD;
}

/* Subtract the `struct timeval' values X and Y,
	 storing the result in RESULT.
	 Return 1 if the difference is negative, otherwise 0.  */

static int
timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y) {
  /* Perform the carry for the later subtraction by updating Y. */
  if (x->tv_usec < y->tv_usec) {
    int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
    y->tv_usec -= 1000000 * nsec;
    y->tv_sec += nsec;
  }
  if (x->tv_usec - y->tv_usec > 1000000) {
    int nsec = (x->tv_usec - y->tv_usec) / 1000000;
    y->tv_usec += 1000000 * nsec;
    y->tv_sec -= nsec;
  }
	
  /* Compute the time remaining to wait.
		 `tv_usec' is certainly positive. */
  result->tv_sec = x->tv_sec - y->tv_sec;
  result->tv_usec = x->tv_usec - y->tv_usec;
     
	/* Return 1 if result is negative. */
  return x->tv_sec < y->tv_sec;
}

static SANE_Status
calculate_transfer_rate (lm9832_Scanner *s, SANE_Word xdpi, SANE_Word ydpi) {
  SANE_Byte register_03;
  //	int i;
  unsigned char scankb1;
  //  unsigned char scankb2;
  long to_read;
  SANE_Status rc = SANE_STATUS_GOOD;
  SANE_Status status;
  unsigned char *buffer;
  struct timeval tv_before;
  struct timeval tv_after;
  struct timeval tv_result;
  long elapsed_time_ms = 0;
  int bytes_read;
  long bytes_read_total;

  assert (s != 0);

  DBG(5, "calculate_transfer_rate\n");
  RETURN_IF_ERROR(motor (s, SANE_FALSE));

  /* Scan data until pause */
  RETURN_IF_ERROR(lamp_on (s));
  RETURN_IF_ERROR(paper_sensor1_off( s));
  setreg (s, 0x07, 0x03);

 retry:
#if 1
  wait_for_pause(s);
#else
  for (i=0; i<50; ++i) {
    scankb1 = getreg (s, 0x01);
    scankb2 = getreg (s, 0x01);
    register_03 = getreg (s, 0x03);
    if (scankb1 == scankb2) {
      DBG(4, "calculate_transfer_rate: image data=%i\n", (int)scankb2);
      register_03 = getreg (s, 0x03);
      DBG(4, "calculate_transfer_rate: reg 0x03=%i\n", (int)register_03);
      if (register_03 & 0x10) {
	break;
      }
    }
    usleep(100L*1000L);
  }
#endif

  /* Check if pause */
  register_03 = getreg (s, 0x03);
  if (register_03 & 0x10) {
    scankb1 = getreg (s, 0x01);
    to_read = scankb1 * 2 * 1024L;
    buffer = malloc(to_read);
    bytes_read_total = 0;
    if (buffer != 0) {
      /* Init timer */
      gettimeofday (&tv_before, NULL);
      do {
	status = read_available_data (s, buffer, &bytes_read, to_read);
	if (status != SANE_STATUS_GOOD) {
	  rc = status;
	  break;
	}
	bytes_read_total += bytes_read;
	gettimeofday (&tv_after, NULL);
	if (0 == timeval_subtract (&tv_result, &tv_after, &tv_before)) {
	  elapsed_time_ms = tv_result.tv_sec*1000 + tv_result.tv_usec/1000;
	  /*	  s->msrd_parm.datalink_bandwidth = bytes_read_total / 
		  (elapsed_time_ms / (double)1000);*/
	} else {
	  /* crossing day boundary? */
	  free(buffer);
	  goto retry;
	}
	elapsed_time_ms = (tv_after.tv_sec - tv_before.tv_sec) * 1000;
	elapsed_time_ms += (tv_after.tv_usec - tv_before.tv_usec) / 1000;
      } while (elapsed_time_ms < 2000);

      /*      DBG(4, "PC Transfer rate = %d bytes/sec. (%ld/%ld)\n",
	  s->msrd_parms.datalink_bandwidth, bytes_read_total,
	  elapsed_time_ms);*/
			
      rc = SANE_STATUS_GOOD;
    } else {
      rc = SANE_STATUS_INVAL;
    }
  } else {
    rc = SANE_STATUS_INVAL;
  }
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0x00));
  RETURN_IF_ERROR(paper_sensor1_on( s));
  RETURN_IF_ERROR(lamp_off (s));

  /* Save some parameters to paramblock */

  DBG(5, "calculate_transfer_rate: succeeded\n");

  return rc;
}

static void
compute_first_gain_offset (int target, int max, int min, int *gain,
			   int *offset, int *max_gain, int *min_offset) 
{
  *gain = (int) 15.0 * (target / (max - min) - 0.933);
  *offset = (int) (-1.0 * min / (512.0 * 0.0195));
  if (*gain >= 32)
    {
      *gain = (int) 15.0 * (target / 3.0 / (max - min) - 0.933);
      *offset = (int) -3.0 * min / (512.0 * 0.0195);
    }
  if (*gain < 0)
    *gain = 0;
  else if (*gain > 63)
    *gain = 63;
  
  if (*offset < -31)
    *offset = -31;
  else if (*offset > 31)
    *offset = 31;

  *max_gain = 63;
  *min_offset = -31;
}

#if 0
static int
compute_gain_offset (int target, int max, int min, int *gain,
		     int *offset, int *max_gain, int *min_offset)
{
  int gain_stable;
  int is_unstable;

  gain_stable = 1;		/* unless the opposite is said */
  is_unstable = 0;

  if (max > target)
    {
      if (*gain > 0)
	{
	  (*gain)--;
	  *max_gain = *gain;
	  gain_stable = 0;
	  is_unstable |= 1;
	}
      else
	{
	  fprintf (stderr, "error: integration time too long.\n");
	  return -1;
	}
    }
  else
    {
      if (*gain < *max_gain)
	{
	  (*gain)++;
	  gain_stable = 0;
	  is_unstable |= 1;
	}
    }

  if (min == 0)
    {
      if (*offset < 31)
	{
	  (*offset)++;
	  if (gain_stable)
	    *min_offset = *offset;
	  is_unstable |= 1;
	}
      else
	{
	  fprintf (stderr, "error: max static has pixel value == 0\n");
	  return -1;
	}
    }
  else
    {
      if (*offset > *min_offset)
	{
	  (*offset)--;
	  is_unstable |= 1;
	}
    }
  return is_unstable;
}
#endif

static int
compute_gain_offset (unsigned int target_min, unsigned int target_low,
		     unsigned int target_high, unsigned int target_max,
		     unsigned int max, unsigned int min, int *gain,
		     int *offset, int *max_gain, int *min_offset)
{
  int gain_stable;
  int is_unstable;

  gain_stable = 1;		/* unless the opposite is said */
  is_unstable = 0;

  if (max > target_max)
    {
      if (*gain > 0)
	{
	  (*gain)--;
	  *max_gain = *gain;
	  gain_stable = 0;
	  is_unstable |= 1;
	}
      else
	{
	  fprintf (stderr, "error: integration time too long.\n");
	  return -1;
	}
    }
  else if (max < target_high)
    {
      if (*gain < *max_gain)
	{
	  (*gain)++;
	  gain_stable = 0;
	  is_unstable |= 1;
	}
    }
  if (min < target_min)
    {
      if (*offset < 31)
	{
	  (*offset)++;
	  if (gain_stable)
	    *min_offset = *offset;
	  is_unstable |= 1;
	}
      else
	{
	  fprintf (stderr, "error: max static has pixel value == 0\n");
	  return -1;
	}
    }
  else if (min > target_low)
    {
      if (*offset > *min_offset)
	{
	  (*offset)--;
	  is_unstable |= 1;
	}
    }
  return is_unstable;
}

#define OFFSET_CODE_SIGN(off) (((off) < 0) ? (-(off) & 0x1f) | 0x20 : (off))
#define OFFSET_DECODE_SIGN(off) (((off) & 0x20) ? -(off & 0x1f) : (off))

static SANE_Status
do_coarse_calibration (lm9832_Scanner *s, struct coarse_t *coarse)
{
  SANE_Status status;
  unsigned char *cal_line;
  unsigned char *cal_line_ptr;
  int cal_line_size;
  /* local scanning params */
  //  int active_pixels_start;
  //  int line_end;
  //  int data_pixels_start;
  //  int data_pixels_end;
  //  int dpd;
  //  int step_size;
  //  int ff_step_size;
  //  char steps_to_reverse;
  char hdpi_div;
  //  char line_rate_color;
  //  int vdpi;			/* vertical dots per inch */
  int hdpi_code;
  int calibrated;
  int first_time;
  
  int red_offset = 0;
  int green_offset = 0;
  int blue_offset = 0;
      
  int red_gain = 1; 
  int green_gain = 1;
  int blue_gain = 1;

  int min_red_offset = -31;
  int min_green_offset = -31;
  int min_blue_offset = -31;
  
  int max_red_gain = 63;
  int max_green_gain = 63;
  int max_blue_gain = 63;

  int max_red;
  int min_red;
  int max_green;
  int min_green;
  int max_blue;
  int min_blue;
  //  static char me[] = "do_coarse_calibration";

  // DBG (DL_CALL_TRACE, "%s\n", me);

  setreg (s, 0x07, 0x00);
  usleep (10 * 1000);
  
  // vdpi = 150;
  hdpi_code = 0;
  hdpi_div = hdpi_mapping[hdpi_code];
  //  active_pixels_start = 0x40;
  //  line_end = 0x2ee0;
  s->mclk_div = 2;
  //  data_pixels_start = 0x40;
  //  data_pixels_end = (int) (data_pixels_start + s->hw_parms.scan_area_width);
  //  data_pixels_end = MIN(data_pixels_end, line_end - 20);

  //  cal_line_size = s->hw_parms.scan_area_width * 3 * 2 + 2;
  cal_line_size = s->bytes_per_line + 2;

  //  setreg (s, 0x1e, HIBYTE(active_pixels_start));
  //  setreg (s, 0x1f, LOBYTE(active_pixels_start));
  //  setreg (s, 0x20, HIBYTE(line_end));
  //  setreg (s, 0x21, LOBYTE(line_end));
  //  setreg (s, 0x22, HIBYTE(data_pixels_start));
  //  setreg (s, 0x23, LOBYTE(data_pixels_start));
  //  setreg (s, 0x24, HIBYTE(data_pixels_end));
  //  setreg (s, 0x25, LOBYTE(data_pixels_end));
  
  //  setreg (s, 0x26,
  //	  PIXEL_RATE_3_CHANNELS |
  //	  GRAY_CHANNEL_RED |
  //	  TR_RED(0) | TR_GREEN(0) | TR_BLUE(0));

  setreg (s, 0x08, (s->mclk_div - 1) * 2);

  // setreg (s, 0x09, hdpi_code | PIXEL_PACKING(3) | DATAMODE(1));
  // setreg (s, 0x0a, 0);		

  setreg (s, 0x38, OFFSET_CODE_SIGN(red_offset));
  setreg (s, 0x39, OFFSET_CODE_SIGN(green_offset));
  setreg (s, 0x3a, OFFSET_CODE_SIGN(blue_offset));
  setreg (s, 0x3b, red_gain);
  setreg (s, 0x3c, green_gain);
  setreg (s, 0x3d, blue_gain);

  // setreg (s, 0x5e, 0x80);

  setreg (s, 0x3e, 0x00);
  setreg (s, 0x3f, 0x00);
  setreg (s, 0x40, 0x40); /* 0x01 According to LM9832 software specs */
  setreg (s, 0x41, 0x00);

  setreg (s, 0x42, 0x20); /* This register has different meaning on LM9830 !!! */

  //  return SANE_STATUS_GOOD;

  //setreg (s, 0x4e, 0x5b - 0x3c); /* max Kb to pause */
  //setreg (s, 0x4f, 0x02);	/* min Kb to resume */

  //line_rate_color = 1;
  //step_size = (vdpi * line_end * line_rate_color) /
  //  (4 * s->hw_parms.motor_full_steps_per_inch);

  //dpd = compute_dpd (s, step_size, line_end); /* 0x0ada; */
#ifdef DEBUG
  //fprintf (stderr, "dpd = %d\n", dpd);
#endif
  //setreg (s, 0x52, HIBYTE(dpd));
  //setreg (s, 0x53, LOBYTE(dpd));

  //setreg (s, 0x46, HIBYTE(step_size));
  //setreg (s, 0x47, LOBYTE(step_size));

  //ff_step_size = compute_fastfeed_step_size (s->hw_parms.crystal_frequency,
  //				     s->mclk_div,
  //				     s->hw_parms.motor_max_speed,
  //				     s->hw_parms.motor_full_steps_per_inch,
  //				     0); /* 0x0190; */
  //setreg (s, 0x48, HIBYTE(ff_step_size));
  //setreg (s, 0x49, LOBYTE(ff_step_size));
  //setreg (s, 0x4b, 0x15);
  //steps_to_reverse = 0x3f;
  //setreg (s, 0x50, steps_to_reverse);
  //setreg (s, 0x51, 0x15);	/* accel profile */

  /* this is to stay the motor stopped */
  //clearbits (s, 0x45, (1 << 4));
  RETURN_IF_ERROR(motor (s, SANE_FALSE));

  // cache_write (s);

  calibrated = 0;
  first_time = 1;
  cal_line = malloc (cal_line_size + 1024);

  do {
    //		unsigned char cmd_reg;

    /* resets the lm983x before start scanning */
    //setreg (s, 0x07, 0x08);
    //do
    //  {
    // setreg (s, 0x07, 0x03);
    // cmd_reg = getreg (s, 0x07);
    //   }
    // while (cmd_reg != 0x03);

    //		RETURN_IF_ERROR(sanei_lm9832_reset (s));
    //		RETURN_IF_ERROR(wait_scanner_ready (s));

    /* Scan until pause, with motor off */
    RETURN_IF_ERROR(lamp_on (s));
    RETURN_IF_ERROR(paper_sensor1_off( s));
    setreg (s, 0x07, 0x03);
    // wait_for_pause(s);

    cal_line_ptr = cal_line;
    status = read_required_data (s, cal_line_ptr, cal_line_size);
    if (status != SANE_STATUS_GOOD)
      return status;

    setreg (s, 0x07, 0x00);
    {
      unsigned int i;

      //      DBG(4, "%d\t%d\t%d\t%d\t%d\t%d\t\n",
      //      cal_line[0], cal_line[1], cal_line[2], 
      //      cal_line[3], cal_line[4], cal_line[5]);

      min_red = max_red = (cal_line[0] * 256 + cal_line[1]) >> 2;
      min_green = max_green = (cal_line[2] * 256 + cal_line[3]) >> 2;
      min_blue = max_blue = (cal_line[4] * 256 + cal_line[5]) >> 2;
      for (i = 6; i < (s->bytes_per_line); i += 6) {
	int value;
	
	//  DBG(4, "%d\t%d\t%d\t%d\t%d\t%d\t\n",
	//  cal_line[i+0], cal_line[i+1], cal_line[i+2], 
	//  cal_line[i+3], cal_line[i+4], cal_line[i+5]);

	value = cal_line[i] * 256 + cal_line[i + 1];
	value >>= 2;
	if (value > max_red)
	  max_red = value;
	value = cal_line[i + 2] * 256 + cal_line[i + 3];
	value >>= 2;
	if (value > max_green)
	  max_green = value;
	value = cal_line[i + 4] * 256 + cal_line[i + 5];
	value >>= 2;
	if (value > max_blue)
	  max_blue = value;
	value = cal_line[i] * 256 + cal_line[i + 1];
	value >>= 2;
	if (value < min_red)
	  min_red = value;
	value = cal_line[i + 2] * 256 + cal_line[i + 3];
	value >>= 2;
	if (value < min_green)
	  min_green = value;
	value = cal_line[i + 4] * 256 + cal_line[i + 5];
	value >>= 2;
	if (value < min_blue)
	  min_blue = value;
      }
      DBG(4, "max_red:%d max_green:%d max_blue:%d\n",
	  max_red, max_green, max_blue);
      DBG(4, "min_red:%d min_green:%d min_blue:%d\n",
	  min_red, min_green, min_blue);
      
      if (first_time)	{
	first_time = 0;
	compute_first_gain_offset (s->target_value_low.red,
				   max_red, min_red,
				   &red_gain, &red_offset,
				   &max_red_gain, &min_red_offset);
	compute_first_gain_offset (s->target_value_low.green,
				   max_green, min_green,
				   &green_gain, &green_offset,
				   &max_green_gain, &min_green_offset);
	compute_first_gain_offset (s->target_value_low.blue,
				   max_blue, min_blue,
				   &blue_gain, &blue_offset,
				   &max_blue_gain, &min_blue_offset);
      } else {
	int retval;

				/* this code should check return value -1 for error */

	retval = compute_gain_offset (s->target_value_min.red,
				      s->target_value_low.red,
				      s->target_value_high.red,
				      s->target_value_max.red,
				      max_red, min_red,
				      &red_gain, &red_offset,
				      &max_red_gain, &min_red_offset);
	if (retval < 0)
	  break;
	retval |= compute_gain_offset (s->target_value_min.green,
				       s->target_value_low.green,
				       s->target_value_high.green,
				       s->target_value_max.green,
				       max_green, min_green,
				       &green_gain, &green_offset,
				       &max_green_gain, &min_green_offset);
	if (retval < 0)
	  break;
	retval |= compute_gain_offset (s->target_value_min.blue,
				       s->target_value_low.blue,
				       s->target_value_high.blue,
				       s->target_value_max.blue,
				       max_blue, min_blue,
				       &blue_gain, &blue_offset,
				       &max_blue_gain, &min_blue_offset);
	if (retval < 0)
	  break;
	calibrated = !retval;
      }

      setreg (s, 0x3b, red_gain);
      setreg (s, 0x3c, green_gain);
      setreg (s, 0x3d, blue_gain);
      
      setreg (s, 0x38, OFFSET_CODE_SIGN(red_offset));
      setreg (s, 0x39, OFFSET_CODE_SIGN(green_offset));
      setreg (s, 0x3a, OFFSET_CODE_SIGN(blue_offset));

      DBG(4, "%d, %d, %d   %d, %d, %d\n", red_gain, green_gain,
	  blue_gain, red_offset, green_offset, blue_offset);
      //      cache_write (s);
    }
  } while (!calibrated);
  coarse->min_red = min_red;
  coarse->min_green = min_green;
  coarse->min_blue = min_blue;
  coarse->max_red = max_red;
  coarse->max_green = max_green;
  coarse->max_blue = max_blue;
  coarse->red_gain = red_gain;
  coarse->green_gain = green_gain;
  coarse->blue_gain = blue_gain;
  coarse->red_offset = red_offset;
  coarse->green_offset = green_offset;
  coarse->blue_offset = blue_offset;

  RETURN_IF_ERROR(paper_sensor1_on( s));
  // RETURN_IF_ERROR(lamp_off (s));

  return SANE_STATUS_GOOD;
}

static SANE_Status
write_offset_coefficient_data(lm9832_Scanner *s, struct coarse_t *coarse)
{
  int i,j;
  int stepsize = 60;
  unsigned char *data;

  DBG(4, "write_offset_coefficient_data\n");

  data = malloc(5460*2);	

  /* Write Red */
  setreg(s, 0x03, 0x00);
  setreg(s, 0x04, 0x00);
  setreg(s, 0x05, 0x00);
  for (i=0; i<(5460*2); i+=2) {
    data[i+0] = HIBYTE(coarse->min_red);
    data[i+1] = LOBYTE(coarse->min_red);
  }
  DBG(4, "R %d", coarse->min_red);
  //	RETURN_IF_ERROR(wait_scanner_ready(s));
  for (i=0; i<(5460*2); i+=stepsize) {
    sanei_lm9832_write (s, 0x06, data+i, stepsize, SANE_FALSE);
  }
  DBG(4, "\n");

  /* Write Green */
  setreg(s, 0x03, 0x04);
  setreg(s, 0x04, 0x00);
  setreg(s, 0x05, 0x00);
  for (i=0; i<(5460*2); i+=2) {
    data[i+0] = HIBYTE(coarse->min_green);
    data[i+1] = LOBYTE(coarse->min_green);
  }
  DBG(4, "G %d", coarse->min_green);
  //	RETURN_IF_ERROR(wait_scanner_ready(s));
  for (i=0; i<(5460*2); i+=stepsize) {
    sanei_lm9832_write (s, 0x06, data+i, stepsize, SANE_FALSE);
  }
  DBG(4, "\n");

  /* Write Blue */
  setreg(s, 0x03, 0x08);
  setreg(s, 0x04, 0x00);
  setreg(s, 0x05, 0x00);
  for (i=0; i<(5460*2); i+=2) {
    data[i+0] = HIBYTE(coarse->min_blue);
    data[i+1] = LOBYTE(coarse->min_blue);
  }
  DBG(4, "B %d", coarse->min_blue);
  //	RETURN_IF_ERROR(wait_scanner_ready(s));
  for (i=0; i<(5460*2); i+=stepsize) {
    sanei_lm9832_write (s, 0x06, data+i, stepsize, SANE_FALSE);
  }
  DBG(4, "\n");

  DBG(4, "write_offset_coefficient_data OK\n");
  return SANE_STATUS_GOOD;
}

#if 0
static int
write_default_offset_gain (HP4200_Scanner *s, char *gain_offset,
			   int size, int color)
{
  char *check_data;
  int retval;

  setreg (s, 0x03, (color << 1) | 1);
  setreg (s, 0x04, DATA_PORT_WRITE);
  setreg (s, 0x05, 0x00);
  pv8630_write_byte (s->fd, 0x06, PV8630_REPPADDRESS);
  pv8630_init_eppbulk_write (s->fd, size);
  write (s->fd, gain_offset, size);

  check_data = malloc (size);
  setreg (s, 0x03, (color << 1) | 1);
  setreg (s, 0x04, DATA_PORT_READ);
  setreg (s, 0x05, 0x00);
  pv8630_write_byte (s->fd, 0x06, PV8630_REPPADDRESS);
  pv8630_init_eppbulk_read (s->fd, size);
  read (s->fd, check_data, size);
  retval = memcmp (gain_offset, check_data, size);
  free (check_data);
  if (retval != 0)
    {
      fprintf (stderr, "error: color %d has bad gain/offset table\n", color);
    }
#ifdef DEBUG
  else
    fprintf (stderr, "color %d gain/offset table is good\n", color);
#endif

  return 0;
}
#endif

static SANE_Status
write_default_offset_gain(lm9832_Scanner *s, char *gain_offset,
			  int size, int color)
{
  int i,j;
  int stepsize = 60;
  char *check_data;
  int retval;

  DBG(4, "write_default_offset_gain\n");

  setreg (s, 0x03, (color << 1) | 1);
  setreg (s, 0x04, 0x00);
  setreg(s, 0x05, 0x00);
  //	RETURN_IF_ERROR(wait_scanner_ready(s));
  for (i=0; i<(5460*2); i+=stepsize) {
    DBG(4, ".");
    sanei_lm9832_write (s, 0x06, gain_offset+i, stepsize, SANE_FALSE);
  }
  DBG(4, "\n");

  DBG(4, "write_default_offset_gain OK\n");
  return SANE_STATUS_GOOD;
}

static int
compute_corr_code (int average, int min_color, int range, int target)
{
  int value;
  int corr_code;

  value = average - min_color;
  if (value > 0)
    corr_code = (int) (range * ((double)target / (double)value - 1.0) + 0.5);
  else
    corr_code = 0;
  if (corr_code < 0)
    corr_code = 0;
  else if (corr_code > 2048)
    corr_code = 0;
  else if (corr_code > 1023)
    corr_code = 1023;
  return corr_code;
}

static SANE_Status
do_fine_calibration (lm9832_Scanner *s, struct coarse_t *coarse)
{
  SANE_Status status;
  unsigned char *cal_line;
  unsigned char *cal_line_ptr;
  int *average;
  char red_gain_offset[5460 * 2 + 1024];
  char green_gain_offset[5460 * 2 + 1024];
  char blue_gain_offset[5460 * 2 + 1024];
  int *corr_red;
  int *corr_green;
  int *corr_blue;
  int registro[30][5460 * 3];
  int cal_line_size;
  /* local scanning params */
  int active_pixels_start;
  int line_end;
  int line_length;
  int data_pixels_start;
  int data_pixels_end;
  int dpd;
  int step_size;
  int ff_step_size;
  char steps_to_reverse;
  char hdpi_div;
  char line_rate_color;
  int vdpi;			/* vertical dots per inch */
  int hdpi_code;
  int calibrated;
  int first_time;
  int lines_to_process;
  
  //  static char me[] = "do_fine_calibration";

  //  DBG (DL_CALL_TRACE, "%s\n", me);

  setreg (s, 0x07,0x00);
  usleep (10 * 1000);
  
  //  vdpi = 150;
  //  hdpi_code = compute_hdpi_code (s->user_parms.horizontal_resolution);
  hdpi_code = 0;

  /* figure out which horizontal divider to use based on the
     calculated horizontal dpi code */
  hdpi_div = hdpi_mapping[hdpi_code];
  //  active_pixels_start = 0x40;
  //  line_end = 0x2ee0;
  //  line_length = s->user_parms.image_width * hdpi_div;
  line_length = s->bytes_per_line;
  s->mclk_div = 2;
  //  data_pixels_start = 0x72 + s->runtime_parms.first_pixel * hdpi_div;
  //  data_pixels_end = (int) (data_pixels_start + s->user_parms.image_width * hdpi_div);
  //  data_pixels_end = min(data_pixels_end, line_end - 20);

  //  cal_line_size = line_length * 3 * 2 + 2;
  cal_line_size = s->bytes_per_line + 2;

  //  setreg (s, 0x1e, HIBYTE(active_pixels_start));
  //  setreg (s, 0x1f, LOBYTE(active_pixels_start));
  //  setreg (s, 0x20, HIBYTE(line_end));
  //  setreg (s, 0x21, LOBYTE(line_end));
  //  setreg (s, 0x22, HIBYTE(data_pixels_start));
  //  setreg (s, 0x23, LOBYTE(data_pixels_start));
  //  setreg (s, 0x24, HIBYTE(data_pixels_end));
  //  setreg (s, 0x25, LOBYTE(data_pixels_end));
  
  //  setreg (s, 0x26,
  //					PIXEL_RATE_3_CHANNELS |
  //					GRAY_CHANNEL_RED |
  //					TR_RED(0) | TR_GREEN(0) | TR_BLUE(0));

  setreg (s, 0x08, (s->mclk_div - 1) * 2);

  //  setreg (s, 0x09, 0 | PIXEL_PACKING(3) | DATAMODE(1));
  //  setreg (s, 0x0a, 0);		

  setreg (s, 0x38, 1);
  setreg (s, 0x39, 1);
  setreg (s, 0x3a, 1);
  setreg (s, 0x3b, coarse->red_gain);
  setreg (s, 0x3c, coarse->green_gain);
  setreg (s, 0x3d, coarse->blue_gain);

  //  setreg (s, 0x5e, 0x80);

  setreg (s, 0x3e, 0x00);
  setreg (s, 0x3f, 0x00);
  setreg (s, 0x40, 0x40); /* 0x01 According to LM9832 software specs */
  setreg (s, 0x41, 0x00);

  write_offset_coefficient_data(s, coarse);

  setreg (s, 0x4a, 0x00);
  setreg (s, 0x4b, 0x00);
  setreg (s, 0x50, 0x00);
  setreg (s, 0x51, 0x00);
  setreg (s, 0x42, 0x24); /* This register has different meaning on LM9830 !!! */

  //  setreg (s, 0x4e, 0x5b - 0x3c); /* max Kb to pause */
  //  setreg (s, 0x4f, 0x02);	/* min Kb to resume */

  //  line_rate_color = 1;
  //  step_size = (vdpi * line_end * line_rate_color) /
  //    (4 * s->hw_parms.motor_full_steps_per_inch);

  //  dpd = compute_dpd (s, step_size, line_end); /* 0x0ada; */
#ifdef DEBUG
  //  fprintf (stderr, "dpd = %d\n", dpd);
#endif
  //  setreg (s, 0x52, HIBYTE(dpd));
  //  setreg (s, 0x53, LOBYTE(dpd));

  //  setreg (s, 0x46, HIBYTE(step_size));
  //  setreg (s, 0x47, LOBYTE(step_size));

  //  ff_step_size = compute_fastfeed_step_size (s->hw_parms.crystal_frequency,
  //																						 s->mclk_div,
  //																						 s->hw_parms.motor_max_speed,
  //																						 s->hw_parms.motor_full_steps_per_inch,
  //																						 0); /* 0x0190; */
  //  setreg (s, 0x48, HIBYTE(ff_step_size));
  //	setreg (s, 0x49, LOBYTE(ff_step_size));
  //  setreg (s, 0x4b, 0x15);
  //  steps_to_reverse = 0x3f;
  //  setreg (s, 0x50, steps_to_reverse);
  //  setreg (s, 0x51, 0x15);	/* accel profile */

  /* this is to activate the motor */
  //  setbits (s, 0x45, (1 << 4));

  lines_to_process = 12; //8 * step_size * 4 / line_end;
  if (lines_to_process < 1)
    lines_to_process = 1;

#ifdef DEBUG
  fprintf (stderr, "lines to process = %d\n", lines_to_process);
#endif

  //  setreg (s, 0x58, 0);

  //  cache_write (s);

  calibrated = 0;
  first_time = 1;
  cal_line = malloc (cal_line_size + 1024);
  average = malloc (sizeof (int) * line_length * 3);
  memset (average, 0, sizeof (int) * line_length * 3);
  { 
    int i;
    for (i = 0; i < 12; i++) {
      memset (registro[i], 0, 5460 * 3);
    }
  }

  /* resets the lm983x before start scanning */
  //  setreg (s, 0x07, 0x08);
  //  setreg (s, 0x07, 0x03);
  RETURN_IF_ERROR(lamp_on (s));
  RETURN_IF_ERROR(paper_sensor1_off( s));
  setreg (s, 0x07, 0x03);
    
  usleep(100);

  do {

    cal_line_ptr = cal_line;

    status = read_required_data (s, cal_line_ptr, cal_line_size);
    if (status != SANE_STATUS_GOOD)
      return status;
    {
      int i, j;

      if (calibrated == 0)
	for (j = 0, i = 0; i < (line_length * 3); i++, j += 2)
	  {
	    average[i] = (cal_line[j] * 256 + cal_line[j + 1]) >> 2;
	    registro[calibrated][i] = average[i];
	    if (i>100 && i<110) DBG(4, "avg[%i]=%i ", i, average[i]);
	  }
      else
	for (j = 0, i = 0; i < (line_length * 3); i++, j += 2)
	  {
	    int value;
	    value = (cal_line[j] * 256 + cal_line[j + 1]) >> 2;
	    average[i] += value;
	    average[i] /= 2;
	    registro[calibrated][i] = average[i];
	    if (i>100 && i<110) DBG(4, "avg[%i]=%i ", i, average[i]);
	  }
    }
    calibrated++;
  } while (calibrated < lines_to_process);
  setreg (s, 0x07, 0x00);
  usleep (10 * 1000);

  /* Restore Acceleration profile */	
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x50, 0x30));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x51, 0xfc));

#if 0
  return SANE_STATUS_GOOD;
#else
#if 0
  {
    int i;
    int j = 0;
    do {
      for (i = 3; (i+6) < (line_length * 3); i += 3)
	{
	  average[i] = (2*average[i-3] + average[i] + 2*average[i+3]) / 5; 
	  average[i+1] = (2*average[i-2] + average[i+1] + 2*average[i+4]) / 5;
	  average[i+2] = (2*average[i-1] + average[i+2] + 2*average[i+5]) / 5;
	}
      j++;
    } while (j < 3);
  }
#endif
  {
    int i;
    int max_red;
    int min_red;
    int max_green;
    int min_green;
    int max_blue;
    int min_blue;
    min_red = max_red = average[0];
    min_green = max_green = average[1];
    min_blue = max_blue = average[2];
    for (i = 3; i < (line_length * 3); i += 3)
      {
	int value;

	value = average[i];
	if (value > max_red)
	  max_red = value;
	value = average[i + 1];
	if (value > max_green)
	  max_green = value;
	value = average[i + 2];
	if (value > max_blue)
	  max_blue = value;
	value = average[i];
	if (value < min_red)
	  min_red = value;
	value = average[i + 1];
	if (value < min_green)
	  min_green = value;
	value = average[i + 2];
	if (value < min_blue)
	  min_blue = value;
      }
    DBG(4, "max_red:%d max_green:%d max_blue:%d\n",
	max_red, max_green, max_blue);
    DBG(4, "min_red:%d min_green:%d min_blue:%d\n",
	min_red, min_green, min_blue);
      
    /* do fine calibration */
    {
      int min_white_red;
      int min_white_green;
      int min_white_blue;
      double ratio;
      int range;
      double aux;
      int min_white_err;
      int j;

      min_white_red = min_white_green = min_white_blue = 0x3ff;
      for (i = 0; i < (line_length * 3); i += 3)
	{
	  int value;

	  value = average[i] - coarse->min_red;
	  if ((value > 0) && (value < min_white_red))
	    min_white_red = value;
	  value = average[i + 1] - coarse->min_green;
	  if ((value > 0) && (value < min_white_green))
	    min_white_green = value;
	  value = average[i + 2] - coarse->min_blue;
	  if ((value > 0) && (value < min_white_blue))
	    min_white_blue = value;
	}

      ratio = 0;
      min_white_err = 0x3ff;

      aux = (double) s->target_value_high.red / min_white_red;
      if (aux > ratio)
	ratio = aux;
      if (min_white_err > min_white_red)
	min_white_err = min_white_red;
      aux = (double) s->target_value_high.green / min_white_green;
      if (aux > ratio)
	ratio = aux;
      if (min_white_err > min_white_green)
	min_white_err = min_white_green;
      aux = (double) s->target_value_high.blue / min_white_blue;
      if (aux > ratio)
	ratio = aux;
      if (min_white_err > min_white_blue)
	min_white_err = min_white_blue;

      DBG(4, "min_white_err = %d, ratio = %f\n", min_white_err,
	  ratio);
      if (ratio <= 1.5)
	range = 2048;
      else if (ratio <= 2.0)
	range = 1024;
      else
	range = 512;

      corr_red = malloc (sizeof(int) * line_length * 3);
      corr_green = malloc (sizeof(int) * line_length * 3);
      corr_blue = malloc (sizeof(int) * line_length * 3);

      for (i = 0, j = 0; i < (line_length * 3); i += 3, j++)
	{
	  corr_red[j] = compute_corr_code (average[i],
					   coarse->min_red,
					   range,
					   s->target_value_high.red);
	  corr_green[j] = compute_corr_code (average[i + 1],
					     coarse->min_green,
					     range,
					     s->target_value_high.green);
	  corr_blue[j] = compute_corr_code (average[i + 2],
					    coarse->min_blue,
					    range,
					    s->target_value_high.blue);
	}
      DBG(4, "After compute_corr\n");
#ifdef DEBUG
      {
	FILE *kaka;
	int i;
	kaka = fopen ("corr.raw", "w");
	for (i = 0; i < line_length; i++)
	  {
	    fprintf (kaka, "%d %d %d %d %d %d ",
		     corr_red[i], corr_green[i], corr_blue[i],
		     average[3 * i], average[3 * i + 1],
		     average[3 * i + 2]);
	    fprintf (kaka, "%d %d %d  %d %d %d  %d %d %d ",
		     registro[0][3*i], registro[0][3*i+1], registro[0][3*i+2], 
		     registro[1][3*i], registro[1][3*i+1], registro[1][3*i+2], 
		     registro[2][3*i], registro[2][3*i+1], registro[2][3*i+2]
		     );
	    fprintf (kaka, "%d %d %d  %d %d %d  %d %d %d\n",
		     registro[3][3*i], registro[3][3*i+1], registro[3][3*i+2], 
		     registro[4][3*i], registro[4][3*i+1], registro[4][3*i+2], 
		     registro[5][3*i], registro[5][3*i+1], registro[5][3*i+2]
		     );
	  }
	fclose (kaka);
      }
#endif
      {
	int max_black;
	int use_six_eight_bits;

	max_black = MAX(coarse->min_red, coarse->min_green);
	max_black = MAX(max_black, coarse->min_blue);
	use_six_eight_bits = (max_black < 64);
	  
	if (use_six_eight_bits)
	  {
	    //setreg (s, 0x3e, (1 << 4) | (1 << 3) | (1024 / range));
	  }
	else
	  {
	    //setreg (s, 0x3e, (1 << 4) | (1 << 3) | (1 << 2) | (1024 /
	    //																									 range));
	  }
	DBG(4, "data_pixels_start, active_pixels_start, line_length: %i, %i, %i\n", 
	    data_pixels_start, active_pixels_start, line_length);
	for (i = 0, j = (data_pixels_start - active_pixels_start) * 2;
	     i < line_length; i++, j += 2)
	  {
	    //						DBG(4, "%i ", j);
	    if (use_six_eight_bits)
	      {
		red_gain_offset[j] = (coarse->min_red << 2) |
		  ((corr_red[i] >> 8) & 0x03);
		red_gain_offset[j + 1] = corr_red[i] & 0xff;
		green_gain_offset[j] = (coarse->min_green << 2) |
		  ((corr_green[i] >> 8) & 0x03);
		green_gain_offset[j + 1] = corr_green[i] & 0xff;
		blue_gain_offset[j] = (coarse->min_blue << 2) |
		  ((corr_blue[i] >> 8) & 0x03);
		blue_gain_offset[j + 1] = corr_blue[i] & 0xff;
	      }
	    else
	      {
		red_gain_offset[j] = coarse->min_red;
		red_gain_offset[j + 1] = corr_red[j] >> 2;
		green_gain_offset[j] = coarse->min_green;
		green_gain_offset[j + 1] = corr_green[j] >> 2;
		blue_gain_offset[j] = coarse->min_blue;
		blue_gain_offset[j + 1] = corr_blue[j] >> 2;
	      }
	  }
	DBG(4, "After compute offset\n");
	write_default_offset_gain (s, red_gain_offset,
				   5460 * 2, 0);
	write_default_offset_gain (s, green_gain_offset,
				   5460 * 2, 1);
	write_default_offset_gain (s, blue_gain_offset,
				   5460 * 2, 2);
      }
    }
  }

  /* turn on offset gain */
  setreg (s, 0x42, 0x26); /* This register has different meaning on LM9830 !!! */

  DBG(4, "do_fine_calibration: OK\n");
  return SANE_STATUS_GOOD;

#endif
}

/* load the RGB gamma tables with a linear profile */
static SANE_Status
set_gamma_table_flat(lm9832_Scanner *s) {
  int colour;
  int offset;
  SANE_Byte buffer[16]; 
  /* for each colour */
  for (colour = 0; colour < 3; colour++) {
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 3, (colour * 4) +2));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 4, 0));
    RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 5, 0));
    for (offset = 0; offset < 256; offset++) {
      memset(buffer,offset,16);
      RETURN_IF_ERROR(sanei_lm9832_write(s,6,buffer,16,0));
    }
  }
  return SANE_STATUS_GOOD;
}

static SANE_Status
add_image_data (lm9832_Scanner *s, FILE *out, unsigned char byte) {
  assert(s != 0);

  s->rrggbb[s->rrggbb_index++] = byte;

  if (s->rrggbb_index >= 6) {
    /* Interleave the bytes from rrggbb to rgbrgb */
    if (s->scan_mode == color14) {
      s->rgbrgb[0] = s->rrggbb[0];
      s->rgbrgb[1] = s->rrggbb[2];
      s->rgbrgb[2] = s->rrggbb[4];
      s->rgbrgb[3] = s->rrggbb[0];
      s->rgbrgb[4] = s->rrggbb[2];
      s->rgbrgb[5] = s->rrggbb[4];
      /* Write to file */
      if (out != 0) {
	fwrite(s->rgbrgb, 3 , 1, out); /* Because rawtoppm only support 8 bit color we only write MSB here */
      }
    } else {
      s->rgbrgb[0] = s->rrggbb[0];
      s->rgbrgb[1] = s->rrggbb[2];
      s->rgbrgb[2] = s->rrggbb[4];
      s->rgbrgb[3] = s->rrggbb[1];
      s->rgbrgb[4] = s->rrggbb[3];
      s->rgbrgb[5] = s->rrggbb[5];
	/* Write to file */
      if (out != 0) {
	fwrite(s->rgbrgb, 6 , 1, out);
      }
    }
    s->rrggbb_index = 0;
  }

  return SANE_STATUS_GOOD;
}

static SANE_Status
start_scan (lm9832_Scanner *s, FILE *out) {
  int i, j;
  SANE_Word xdpi = 100;
  SANE_Word ydpi = 100;
  SANE_Word start_line = 1000;
  SANE_Word end_line = 5000;
  SANE_Byte register_byte[1];
  SANE_Word color_mode = /*color8;*/ color14;
  unsigned char buffer[16384];
  unsigned int bytes_read;
  unsigned int bytes_written;
  unsigned int lines_written;
  unsigned int stalled;
  unsigned long tot_bytes_written;
  unsigned int skip_bytes;
  struct coarse_t coarse;

  assert(s != 0);

  DBG(4, "start_scan: fd=%d\n", s->fd);
  RETURN_IF_ERROR(wait_scanner_ready (s));
  RETURN_IF_ERROR(home_sensor (s));
  RETURN_IF_ERROR(wait_scanner_ready (s));
  RETURN_IF_ERROR(move_lines_ff (s, 100)); /* Move to white calibration strip */
  RETURN_IF_ERROR(wait_scanner_ready (s));
  RETURN_IF_ERROR(set_parameters (s, color_mode,
				  xdpi, /* DPI */
				  start_line, /* Start line */
				  0,  /* Start col (600DPI steps) */
				  end_line,	/* End line */
				  10200,  /* End col (600DPI steps) */
				  SANE_FALSE /* No preview */
				  ));
  DBG(4, "#1\n");
  RETURN_IF_ERROR(calculate_transfer_rate (s, xdpi, ydpi));
  DBG(4, "#2\n");
  RETURN_IF_ERROR(print_registers (s));
  //RETURN_IF_ERROR(do_coarse_calibration (s, &coarse));
  DBG(4, "#3\n");
  //RETURN_IF_ERROR(do_fine_calibration (s, &coarse));
  DBG(4, "#4\n");
  setreg(s, 0x28, 0x00); /* Maybe not needed */

  //	setreg(s, 0x38, 0x00);
  //	setreg(s, 0x39, 0x00);
  //	setreg(s, 0x3a, 0x00);

  //	setreg(s, 0x3b, 0x20);
  //	setreg(s, 0x3c, 0x20);
  //	setreg(s, 0x3d, 0x20);

  RETURN_IF_ERROR(print_registers (s));
  DBG(4, "#5\n");
  RETURN_IF_ERROR(wait_scanner_ready (s));
  DBG(4, "#6\n");

  //	return SANE_STATUS_GOOD;

  RETURN_IF_ERROR(motor( s, SANE_TRUE));
  RETURN_IF_ERROR(wait_scanner_ready (s));
  RETURN_IF_ERROR(move_lines_ff (s, start_line));
  RETURN_IF_ERROR(wait_scanner_ready (s));
  RETURN_IF_ERROR(print_registers (s));

  RETURN_IF_ERROR(sanei_lm9832_read (s, 1, register_byte,
				     1, SANE_TRUE));
  DBG(4, "start_scan: image data=%i\n", (int)register_byte[0]);
  RETURN_IF_ERROR(lamp_on (s));
  RETURN_IF_ERROR(paper_sensor1_off( s));
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0x03));

  RETURN_IF_ERROR(sanei_lm9832_read (s, 1, register_byte,
				     1, SANE_TRUE));
  DBG(4, "start_scan: image data=%i\n", (int)register_byte[0]);

  bytes_read = 0;
  bytes_written = 0;
  lines_written = 0;
  tot_bytes_written = 0;
  skip_bytes = 0;
  stalled = 0;

  // for (i=0; i<250; ++i) {
  do {
    read_available_data (s, buffer, &bytes_read, sizeof(buffer));	
    if (bytes_read == 0) {
      stalled++;
    } else {
      stalled = 0;
    }
    // read_required_data (s, buffer, s->bytes_per_line+2); 
    // bytes_read = s->bytes_per_line+2;
    /* Write to out stream. Format is rrggbbrrggbb..... */
    DBG(5, "start_scan: bytes_read=%i\n", (int)bytes_read);
    for (j=0; j<bytes_read; ++j) {
      if (skip_bytes > 0) {
	skip_bytes -= 1;
      } else {
	add_image_data (s, out, buffer[j]);
	bytes_written += 1;
	tot_bytes_written += 1;
	if (bytes_written >= (s->bytes_per_line)) {
	  DBG(5, "start_scan: eol. bytes written=%i tot written=%li\n", 
	      (int)bytes_written, (long)tot_bytes_written);
	  skip_bytes += 2;
	  bytes_written = 0;
	  lines_written += 1;						
	}
      }
    }
  } while (lines_written < (end_line - start_line) && stalled < 50);

  /* Stop scan */
  RETURN_IF_ERROR(sanei_lm9832_write_byte (s, 0x07, 0x00));

  //	usleep (4 * 1000 * 1000); 
  RETURN_IF_ERROR(wait_scanner_ready (s));
  RETURN_IF_ERROR(paper_sensor1_on( s));
  RETURN_IF_ERROR(home_sensor (s));
  RETURN_IF_ERROR(wait_scanner_ready (s));
  RETURN_IF_ERROR(motor( s, 0));
  
  DBG(4, "start_scan: succeded\n");

  return SANE_STATUS_GOOD;
}

/*** Functions for playing a little with the scanner ***/

/* Move fast forward and reverse */
static SANE_Status
play_move_ff_fr(lm9832_Scanner *s) {
  assert(s != 0);

  RETURN_IF_ERROR(wait_scanner_ready (s));
  /* Move forward scan slider for 3000 lines */
  RETURN_IF_ERROR(move_lines_ff (s, 3000));
  usleep (1 * 1000 * 1000); /* Wait a sec */
  /* Reverse scan slider for 500 lines 
     Should be lower then previous forward */
  RETURN_IF_ERROR(move_lines_fr (s, 500)); 
  usleep (1 * 1000 * 1000); /* Wait a sec */
  /* Home sensor */
  RETURN_IF_ERROR(home_sensor (s));
  RETURN_IF_ERROR(wait_scanner_ready (s));

  return SANE_STATUS_GOOD;
}

/* Move fast forward and reverse */
static SANE_Status
play_move_ff_secs(lm9832_Scanner *s) {
  assert(s != 0);

  RETURN_IF_ERROR(wait_scanner_ready (s));
  /* Move forward scan slider for 2 seconds */
  RETURN_IF_ERROR(move_secs_ff (s, 2));

  /* Home sensor */
  RETURN_IF_ERROR(home_sensor (s));
  RETURN_IF_ERROR(wait_scanner_ready (s));

  return SANE_STATUS_GOOD;
}


/***************************************************************************/
/* don't have main procedure if building for SANE */
#ifdef TEST
int
main(void) 
{
  SANE_Int fd;
  SANE_Bool home;
  SANE_Byte version;
  lm9832_Scanner scanner;
  lm9832_Scanner *s;


  s = &scanner; /* Until param is passed at all places */

  memset(&scanner, 0, sizeof(scanner));

  /*  RETURN_IF_ERROR(sanei_lm9832_open (s, "/dev/usbscanner", 8192));*/
  sane_init(NULL,NULL);
  RETURN_IF_ERROR(sane_open("/dev/usbscanner",s));
  fd = scanner.fd; /* Until s param is passed at all places */
  RETURN_IF_ERROR(sanei_lm9832_detect (s, &version));
  if ((version < 3) || (version > 4))
    {
      DBG(1, "main: This is not a LM9832 or 9832 chip based scanner.\n");
      return SANE_STATUS_INVAL;
    }

  RETURN_IF_ERROR(sanei_lm9832_reset (s));
  RETURN_IF_ERROR(led (s, SANE_TRUE));
  RETURN_IF_ERROR(print_registers (s));
  RETURN_IF_ERROR(power (s, SANE_TRUE));
  //	RETURN_IF_ERROR(lamp_on (s));

  /* RETURN_IF_ERROR(play_move_ff_fr (s)); */
  //	RETURN_IF_ERROR(play_move_ff_secs (s));
  {
    SANE_Int i;
    for (i = 0; i < 10 /* Change to non zero to flash led on scanner 
			  that has one ???
			  hp scanjet 2200c does not */
	   ; i++)
      {
	RETURN_IF_ERROR(led (s, SANE_TRUE));
	usleep (200000);
	RETURN_IF_ERROR(led (s, SANE_FALSE));
	usleep (200000);
      }
    RETURN_IF_ERROR(led (s, SANE_TRUE));
  }

  RETURN_IF_ERROR(check_sensor_home (s, &home));
  RETURN_IF_ERROR(wait_scanner_ready (s));
  RETURN_IF_ERROR(check_sensor_home (s, &home));
  //	RETURN_IF_ERROR(set_parameters (s, color8, 100, 0, 0, 50, 50));
  RETURN_IF_ERROR(print_registers (s));

  RETURN_IF_ERROR(start_scan (s, stdout));

  RETURN_IF_ERROR(lamp_off (s));
  RETURN_IF_ERROR(power (s, SANE_FALSE));
  RETURN_IF_ERROR(led (s, SANE_FALSE));
  //	RETURN_IF_ERROR(sanei_lm9832_reset (fd));
  RETURN_IF_ERROR(sanei_lm9832_close (s));

  return SANE_STATUS_GOOD;
}
#endif

/* stuff fettled from s9036 driver and SANE interface */
static
SANE_Status attach(const char* devname, lm9832_Device** devp){
  lm9832_Device *dev;
  lm9832_Scanner dummy; /* just used for detect portion... */
  SANE_Int fd;
  SANE_Byte version;
  for (dev = sane_lm9832_devices; dev; dev = dev->next)
    if (strcmp (dev->sane.name, devname) == 0)
      {
	if (devp)
	  *devp = dev;
	return SANE_STATUS_GOOD;
      }
  DBG(3,"attach: opening %s\n", devname);
  fd = open(devname,O_RDWR | O_EXCL);
  if (fd < 0) {
    DBG (1, "attach: open failed (%s)\n", strerror(errno));
    return SANE_STATUS_INVAL;
  } else {
    /* check to see if contains a LM9832 chip*/
    DBG(4,"Detecting\n");
    dummy.fd = fd;
    RETURN_IF_ERROR(sanei_lm9832_detect(&dummy,&version));
    DBG(4,"Detected\n");
    close(fd);
    if (version!=4) return SANE_STATUS_INVAL;
    dev = malloc (sizeof (*dev));
    if (!dev)
      return SANE_STATUS_NO_MEM;
    // FIXME memset(dev,0,sizeof(*dev));
    /* FIXME detect scanner type here? */
#ifdef KLUDGE_EPSON_1250
    dev->scanner_make = LM9832_EPSON1250;
#else
    dev->scanner_make = LM9832_HP2200;
#endif
    dev->sane.name = strdup (devname);
    dev->sane.vendor = sane_scanner_details[dev->scanner_make].vendor;
    dev->sane.model = sane_scanner_details[dev->scanner_make].model;
    dev->sane.type = sane_scanner_details[dev->scanner_make].type;
    dev->handle = 0;
    DBG (4, "attach: found %s %s\n",dev->sane.vendor,dev->sane.model);
    ++num_devices;
    dev->next = sane_lm9832_devices;
    sane_lm9832_devices = dev;
    if (devp)
      *devp = dev;    
    return SANE_STATUS_GOOD;
  }
}
static SANE_Bool lm9832_test_little_endian(void)
{
  SANE_Int testvalue = 255;
  unsigned char *firstbyte = (unsigned char *) &testvalue;

  if (*firstbyte == 255)
  {
    return SANE_TRUE;
  }

  return SANE_FALSE;
}

/* calibrate for epson1250 - assumes  scanner is @ home */
static
SANE_Status phils_calibration_bit(lm9832_Scanner* s, SANE_Word pixels) {
  SANE_Byte* white_buffer; /* white values... */
  SANE_Byte* gain_buffer[3]; /* colour gain values... */
  SANE_Byte step_lb; /* storage values for 0x4a & 0x4b*/
  SANE_Byte step_hb;
  SANE_Byte reg09; 
  int colour,a; 
  white_buffer = (SANE_Byte*)(malloc(pixels*6)); /* allocate space for raw data*/
  gain_buffer[0] = (SANE_Byte*)(malloc(pixels*2));
  gain_buffer[1] = (SANE_Byte*)(malloc(pixels*2));
  gain_buffer[2] = (SANE_Byte*)(malloc(pixels*2));
  RETURN_IF_ERROR(sanei_lm9832_read_byte(s,0x4a,&step_hb));
  RETURN_IF_ERROR(sanei_lm9832_read_byte(s,0x4b,&step_lb));
  DBG(5,"CAL2: %#x %#x\n",step_hb,step_lb);
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x4a,0));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x4b,0));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x42,0x20));
  RETURN_IF_ERROR(motor(s,SANE_FALSE)); /* turn motor off for time being - don't want to move head */
  // set up for scanning white area...
  RETURN_IF_ERROR(sanei_lm9832_read_byte(s,0x09,&reg09));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x09,reg09 | 0x20));
  //scan it
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x40,0x10));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x41,0));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x07,3));
  RETURN_IF_ERROR(read_required_data(s,white_buffer,pixels*6));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x07,0));
  /* calculate gain values */
  for (a = 0; a < pixels*3; a++) {
    double temp;
    SANE_Word temp2;
    colour = a % 3; /* r,g,b,r,g,b,r,g,b,.... */
    temp = white_buffer[a*2]*0x100 + white_buffer[a*2+1];
    temp2 = temp;
    DBG(5,"Cal: %#x ",temp2);
    temp = 0xffff * 0x1000 / temp; 
    temp2 = MIN(temp,0xffff);
    DBG(5,"%#x\n",temp2);
    gain_buffer[colour][((a-colour)/3)*2] = HIBYTE(temp2);
    gain_buffer[colour][1+((a-colour)/3)*2] = LOBYTE(temp2);
  }
  RETURN_IF_ERROR(sanei_lm9832_reset(s));
  /* apply calibrations to scanner */
  for (colour = 0; colour < 3; colour++) {
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x03,(colour << 2) +1));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x04,0));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x05,0));
  RETURN_IF_ERROR(sanei_lm9832_write(s,0x06,gain_buffer[colour],pixels*2,0));
  }
  // think, then apply calibrations...
  RETURN_IF_ERROR(motor(s,SANE_TRUE)); /* turn motor on again */
  DBG(5,"CAL2: %#x %#x\n",step_hb,step_lb);
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x4a,step_hb));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x4b,step_lb));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x42,0x22));
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x09,reg09));
  /* free buffers*/
  free(white_buffer);
  free(gain_buffer[0]);
  free(gain_buffer[1]);
  free(gain_buffer[2]);
  return SANE_STATUS_GOOD;
}


static
SANE_Status  make_little_endian(SANE_Byte *buf,SANE_Int len) {
  int a;
  SANE_Byte temp;
  DBG(3,"Making little_endian\n");
  if ((len%2) == 1) return SANE_STATUS_INVAL;
  for (a=0; a< len; a+=2) {
    temp = buf[a];
    buf[a] = buf[a+1];
    buf[a+1] = temp;
  }
  return SANE_STATUS_GOOD;
}
/* uses red and green fifo buffers to realign colour planes */ 
static 
void interleave(lm9832_Scanner *s, SANE_Byte *buf, SANE_Int len) { 
  // work out which is first colour up...
  unsigned char temp;
  int a;
  switch (s->val[OPT_DEPTH]) {
  case 8:
    for (a = 0; a < len; a++) { /* for each byte */
      switch (interleave_colour) {
      case 0: /* ignore red*/
	temp = buf[a];
	buf[a] = s->red_buffer[s->red_buf_count];
	s->red_buffer[s->red_buf_count] = temp;
	if (s->red_buf_count >= (s->params.pixels_per_line * s->colour_buffer_lines*2)) {
	  s->red_buf_count = 0;
	} else {
	  s->red_buf_count++;
	}
	break;
      case 1:
	temp = buf[a];
	buf[a] = s->green_buffer[s->green_buf_count];
	s->green_buffer[s->green_buf_count] = temp;
	if (s->green_buf_count >= (s->params.pixels_per_line * s->colour_buffer_lines)) {
	  s->green_buf_count = 0;
	} else {
	  s->green_buf_count++;
	}	  
	break;
      case 2: /* ignore blue */
	break;
      }
      interleave_colour = (interleave_colour+1)%3;
    }
    break;
  case 16:
    for (a = 0; a < len; a+=2) { /* for each word */
      switch (interleave_colour) {
      case 0: /* ignore red*/
	temp = buf[a];
	buf[a] = s->red_buffer[s->red_buf_count];
	s->red_buffer[s->red_buf_count] = temp;
	temp = buf[a+1];
	buf[a+1] = s->red_buffer[s->red_buf_count+1];
	s->red_buffer[s->red_buf_count+1] = temp;
	if (s->red_buf_count >= (s->params.pixels_per_line * s->colour_buffer_lines*4)) {
	  s->red_buf_count = 0;
	} else {
	  s->red_buf_count+=2;
	}	 
	break;
      case 1:
	temp = buf[a];
	buf[a] = s->green_buffer[s->green_buf_count];
	s->green_buffer[s->green_buf_count] = temp;
	temp = buf[a+1];
	buf[a+1] = s->green_buffer[s->green_buf_count+1];
	s->green_buffer[s->green_buf_count+1] = temp;
	if (s->green_buf_count >= (s->params.pixels_per_line * s->colour_buffer_lines*2)) {
	  s->green_buf_count = 0;
	} else {
	  s->green_buf_count+=2;
	}	  
	break;
      case 2: /* ignore blue */
	break;
      }
      interleave_colour = (interleave_colour+1)%3;
    }
    break;
  }
}


static
void do_cancel(lm9832_Scanner* s) {
  if (s->scanning) {
    s->scanning = 0;
    home_sensor(s);
  }
}

static 
void init_options(lm9832_Scanner* s){
  int i;
  /* Hardware Limitations: must be static ! */
  static const SANE_Int depth_list[] = 
  {3,1,8,16};

  memset (s->opt,0,sizeof(s->opt));
  memset (s->val,0,sizeof(s->val));

  for (i = 0; i < NUM_OPTIONS; ++i)
    {
      s->opt[i].size = sizeof (SANE_Word);
      s->opt[i].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    }

  s->opt[OPT_NUM_OPTS].title = SANE_TITLE_NUM_OPTIONS;
  s->opt[OPT_NUM_OPTS].desc = SANE_DESC_NUM_OPTIONS;
  s->opt[OPT_NUM_OPTS].cap = SANE_CAP_SOFT_DETECT;
  s->val[OPT_NUM_OPTS] = NUM_OPTIONS;

  /* "Mode" group: */

  s->opt[OPT_MODE_GROUP].title = "Scan Mode";
  s->opt[OPT_MODE_GROUP].desc = "";
  s->opt[OPT_MODE_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_MODE_GROUP].cap = 0;
  s->opt[OPT_MODE_GROUP].constraint_type = SANE_CONSTRAINT_NONE;



  s->opt[OPT_DEPTH].name = SANE_NAME_BIT_DEPTH;
  s->opt[OPT_DEPTH].title = SANE_TITLE_BIT_DEPTH;
  s->opt[OPT_DEPTH].desc = SANE_DESC_BIT_DEPTH;
  s->opt[OPT_DEPTH].type = SANE_TYPE_INT;
  s->opt[OPT_DEPTH].unit = SANE_UNIT_BIT;
  s->opt[OPT_DEPTH].size = sizeof(SANE_Word);
  s->opt[OPT_DEPTH].constraint_type = SANE_CONSTRAINT_WORD_LIST;
  s->opt[OPT_DEPTH].constraint.word_list = depth_list;
  s->val[OPT_DEPTH] = 8;

  /* resolution */
  s->opt[OPT_RESOLUTION].name = SANE_NAME_SCAN_RESOLUTION;
  s->opt[OPT_RESOLUTION].title = SANE_TITLE_SCAN_RESOLUTION;
  s->opt[OPT_RESOLUTION].desc = SANE_DESC_SCAN_RESOLUTION;
  s->opt[OPT_RESOLUTION].type = SANE_TYPE_INT;
  s->opt[OPT_RESOLUTION].unit = SANE_UNIT_DPI;
  s->opt[OPT_RESOLUTION].size = sizeof(SANE_Word);
  s->opt[OPT_RESOLUTION].constraint_type = SANE_CONSTRAINT_WORD_LIST;
  s->opt[OPT_RESOLUTION].constraint.word_list = s->details->dpi_list;
  s->val[OPT_RESOLUTION] = s->details->dpi_list[1];

  /* preview ; not currently used*/
  s->opt[OPT_PREVIEW].name = SANE_NAME_PREVIEW;
  s->opt[OPT_PREVIEW].title = SANE_TITLE_PREVIEW;
  s->opt[OPT_PREVIEW].desc = SANE_DESC_PREVIEW;
  s->opt[OPT_PREVIEW].type = SANE_TYPE_BOOL;
  s->opt[OPT_PREVIEW].unit = SANE_UNIT_NONE;
  s->opt[OPT_PREVIEW].size = sizeof(SANE_Word);
  s->val[OPT_PREVIEW] = 0;
#if 0    
  s->opt[OPT_CALIBRATE].name = "calibrate";
  s->opt[OPT_CALIBRATE].title = "Calibrate";
  s->opt[OPT_CALIBRATE].desc = "Calibrate before scan?";
  s->opt[OPT_CALIBRATE].type = SANE_TYPE_BOOL;
  s->opt[OPT_CALIBRATE].unit = SANE_UNIT_NONE;
  s->opt[OPT_CALIBRATE].size = sizeof(SANE_Word);
  s->val[OPT_CALIBRATE] = SANE_TRUE;
#endif    
  /* "Geometry" group: */
  s->opt[OPT_GEOMETRY_GROUP].title = "Geometry";
  s->opt[OPT_GEOMETRY_GROUP].desc = "";
  s->opt[OPT_GEOMETRY_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_GEOMETRY_GROUP].cap = SANE_CAP_ADVANCED;
  s->opt[OPT_GEOMETRY_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

  /* top-left x */
  s->opt[OPT_TL_X].name = SANE_NAME_SCAN_TL_X;
  s->opt[OPT_TL_X].title = SANE_TITLE_SCAN_TL_X;
  s->opt[OPT_TL_X].desc = SANE_DESC_SCAN_TL_X;
  s->opt[OPT_TL_X].type = SANE_TYPE_FIXED;
  s->opt[OPT_TL_X].unit = SANE_UNIT_MM;
  s->opt[OPT_TL_X].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_TL_X].constraint.range = &s->details->x_range;
  s->val[OPT_TL_X] = 0;

  /* top-left y */
  s->opt[OPT_TL_Y].name = SANE_NAME_SCAN_TL_Y;
  s->opt[OPT_TL_Y].title = SANE_TITLE_SCAN_TL_Y;
  s->opt[OPT_TL_Y].desc = SANE_DESC_SCAN_TL_Y;
  s->opt[OPT_TL_Y].type = SANE_TYPE_FIXED;
  s->opt[OPT_TL_Y].unit = SANE_UNIT_MM;
  s->opt[OPT_TL_Y].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_TL_Y].constraint.range = &s->details->y_range;
  s->val[OPT_TL_Y] = 0;

  /* bottom-right x */
  s->opt[OPT_BR_X].name = SANE_NAME_SCAN_BR_X;
  s->opt[OPT_BR_X].title = SANE_TITLE_SCAN_BR_X;
  s->opt[OPT_BR_X].desc = SANE_DESC_SCAN_BR_X;
  s->opt[OPT_BR_X].type = SANE_TYPE_FIXED;
  s->opt[OPT_BR_X].unit = SANE_UNIT_MM;
  s->opt[OPT_BR_X].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BR_X].constraint.range = &s->details->x_range;
  s->val[OPT_BR_X] = s->details->x_range.max;

  /* bottom-right y */
  s->opt[OPT_BR_Y].name = SANE_NAME_SCAN_BR_Y;
  s->opt[OPT_BR_Y].title = SANE_TITLE_SCAN_BR_Y;
  s->opt[OPT_BR_Y].desc = SANE_DESC_SCAN_BR_Y;
  s->opt[OPT_BR_Y].type = SANE_TYPE_FIXED;
  s->opt[OPT_BR_Y].unit = SANE_UNIT_MM;
  s->opt[OPT_BR_Y].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BR_Y].constraint.range =  &s->details->y_range;
  s->val[OPT_BR_Y] = s->details->y_range.max;
}

SANE_Status sane_init (SANE_Int * version_code,
		       SANE_Auth_Callback authorize){
  char dev_name[PATH_MAX];
  size_t len;
  FILE *fp;
  DBG(1,"Starting\n");
#ifndef TEST
  if (version_code) *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, 0);
#endif
  little_endian = lm9832_test_little_endian();
  /* FIXME: ignore authorize */
  fp = sanei_config_open ("lm9832.conf");
  if (!fp)
    {
      /* default to /dev/scanner instead of insisting on config file */
      attach ("/dev/scanner", 0);
      return SANE_STATUS_GOOD;
    }

  while (sanei_config_read (dev_name, sizeof (dev_name), fp))
    {
      if (dev_name[0] == '#')	/* ignore line comments */
	continue;
      len = strlen (dev_name);

      if (!len)
	continue;		/* ignore empty lines */

      attach(dev_name,0);
    }
  fclose (fp);
  return SANE_STATUS_GOOD;
}

void sane_exit (void){
  lm9832_Device *dev, *next;
  DBG(3,"Exiting\n");
  for (dev = sane_lm9832_devices; dev; dev = next)
    {
      next = dev->next;
      if (dev->handle)
	sane_close (dev->handle);
      free(dev);
    }
}
SANE_Status
sane_get_devices (const SANE_Device *** device_list, SANE_Bool local_only)
{
  static const SANE_Device **devlist = 0;
  lm9832_Device *dev;
  int i;

  DBG(3,"Getting devices\n");
  if (devlist)
    free (devlist);

  devlist = malloc ((num_devices + 1) * sizeof (devlist[0]));
  if (!devlist)
    return SANE_STATUS_NO_MEM;
  dev = sane_lm9832_devices;
  for (i = 0; i < num_devices; dev = dev->next)
    devlist[i++] = &dev->sane;
  devlist[i++] = 0;

  *device_list = devlist;
  DBG(3,"Got devices\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_open (SANE_String_Const devicename, SANE_Handle * handle)
{
  lm9832_Device *dev;
  lm9832_Scanner *s;
  SANE_Int fd;
  
  DBG(3,"opening\n");
  if (devicename[0]) {
    RETURN_IF_ERROR(attach(devicename,&dev));
  } else {
    dev = sane_lm9832_devices;
  }
  if (!dev) return SANE_STATUS_INVAL;
  if (dev->handle)
    return SANE_STATUS_DEVICE_BUSY;

  s = malloc (sizeof (*s));
  if (!s)
    return SANE_STATUS_NO_MEM;

  memset (s, 0, sizeof (*s));
  s->scanning = 0;
  s->hw = dev;
  s->hw->handle = s;
  s->details = &sane_scanner_details[s->hw->scanner_make];
  RETURN_IF_ERROR(sanei_lm9832_open(s,s->hw->sane.name,8096));
  init_options (s);
  power(s,SANE_TRUE); /* turn scanner on, home sensor, turn lamp on*/
  lamp_on(s); /* make sure... */
  s->blocking = 1;
  s->timer = time(NULL);
  s->red_buffer = NULL; /* set to NULL if freed. Only free on close or on start*/
  s->green_buffer = NULL; /* set to NULL if freed. Only free on close or on start*/
  *handle = s;
  DBG(3,"Opened\n");
  return SANE_STATUS_GOOD;
}

void
sane_close (SANE_Handle handle)
{
  SANE_Int fd;
  lm9832_Scanner *s = handle;
  DBG(3,"Closing\n");
  if (s->scanning)
    do_cancel (handle);
  lamp_off(s);
  power(s,SANE_FALSE);
  sanei_lm9832_close(s);
  if (s->red_buffer) free(s->red_buffer);
  if (s->green_buffer) free(s->green_buffer);
  s->hw->handle = NULL;
  free (handle);
}

const SANE_Option_Descriptor *
sane_get_option_descriptor (SANE_Handle handle, SANE_Int option)
{
  lm9832_Scanner *s = handle;
  if ((unsigned) option >= NUM_OPTIONS)
    return 0;
  return s->opt + option;
}

SANE_Status
sane_control_option (SANE_Handle handle, SANE_Int option,
		     SANE_Action action, void *val, SANE_Int * info)
{
  lm9832_Scanner *s = handle;
  SANE_Status status;
  DBG(5,"Probing option %d\n",option);
  if (info)
    *info = 0;

  if (s->scanning)
    return SANE_STATUS_DEVICE_BUSY;

  if (option >= NUM_OPTIONS || !SANE_OPTION_IS_ACTIVE (s->opt[option].cap))
    return SANE_STATUS_UNSUPPORTED;

  if (action == SANE_ACTION_GET_VALUE)
    {

      switch (option)
	{
	case OPT_DEPTH:
	case OPT_RESOLUTION:
	case OPT_PREVIEW:
#if 0
	case OPT_CALIBRATE:
#endif
	case OPT_TL_X:
	case OPT_TL_Y:
	case OPT_BR_X:
	case OPT_BR_Y:
	case OPT_NUM_OPTS:
	  *(SANE_Word *) val = s->val[option];
	  break;
	default:
	  return SANE_STATUS_UNSUPPORTED;
	}

    }
  else if (action == SANE_ACTION_SET_VALUE)
    {

      if (!SANE_OPTION_IS_SETTABLE (s->opt[option].cap))
	return SANE_STATUS_UNSUPPORTED;

      status = sanei_constrain_value (s->opt + option, val, info);
      if (status != SANE_STATUS_GOOD)
	return status;

      switch (option)
	{
	case OPT_DEPTH:
	case OPT_RESOLUTION:
	case OPT_PREVIEW:
#if 0
	case OPT_CALIBRATE:
#endif
	case OPT_TL_X:
	case OPT_TL_Y:
	case OPT_BR_X:
	case OPT_BR_Y:
	  if (info)
	    *info |= SANE_INFO_RELOAD_PARAMS;
	  s->val[option] = *(SANE_Word *) val;
	  break;
	default:
	  return SANE_STATUS_UNSUPPORTED;
	}

    }
  else if (action == SANE_ACTION_SET_AUTO)
    {

      if (!SANE_OPTION_IS_SETTABLE (s->opt[option].cap))
	return SANE_STATUS_UNSUPPORTED;

      switch (option)
	{
	default:
	  return SANE_STATUS_UNSUPPORTED;
	}

    }
  else
    return SANE_STATUS_UNSUPPORTED;
  DBG(5,"Option probed\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_get_parameters (SANE_Handle handle, SANE_Parameters * params)
{
  lm9832_Scanner *s = handle;
  DBG(3,"Getting Params %d\n",s->scanning);
  if (1)
    {
      double width, height, dpi;

      memset (&s->params, 0, sizeof (s->params));

      s->params.format = SANE_FRAME_RGB;
      s->params.last_frame = SANE_TRUE;

      s->params.depth = s->val[OPT_DEPTH];

      width = SANE_UNFIX (s->val[OPT_BR_X] - s->val[OPT_TL_X]);
      height = SANE_UNFIX (s->val[OPT_BR_Y] - s->val[OPT_TL_Y]);
      dpi = s->val[OPT_RESOLUTION];

      /* make best-effort guess at what parameters will look like once
	 scanning starts.  */
      if (dpi > 0.0 && width > 0.0 && height > 0.0)
	{
	  double dots_per_mm = dpi / MM_PER_INCH;
	  /* we need to arrange that the number of pixels per line */
	  /* results in an even number of bytes per line*/
	  s->params.pixels_per_line = width * dots_per_mm + 0.5;
	  switch (s->params.depth) {
	  case 1:
	    s->params.pixels_per_line -= s->params.pixels_per_line % 16;// make divisible by 16
	    break;
	  case 8:
	    s->params.pixels_per_line -= s->params.pixels_per_line % 2; // make divisible by 2;
	    break;
	  }
	  s->params.lines = height * dots_per_mm + 0.5;
	}
      s->params.bytes_per_line = (s->params.depth * 3 * s->params.pixels_per_line) / 8;
      DBG(3,"Params: %d %d %d\n",s->params.bytes_per_line,s->params.pixels_per_line,s->params.lines);
    }
  if (params)
    *params = s->params;
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_start (SANE_Handle handle)
{
  double lamp_time;
  double temp;
  struct coarse_t coarse;
  lm9832_Scanner *s = handle;
  SANE_Int start_col,end_col,start_line,end_line; /* image geometry in handy units */
  if (s->scanning)
    do_cancel (s);
  sane_get_parameters(handle,NULL);
  home_sensor(s); /* home scanner */
  temp = ceil(s->val[OPT_RESOLUTION]*1.0f)/(1/s->details->sensor_separation);
  s->colour_buffer_lines = temp;
  DBG(3,"Colour_lines = %g, %ld\n",s->details->sensor_separation,s->colour_buffer_lines);
  if (s->red_buffer) free(s->red_buffer);
  if (s->green_buffer) free(s->green_buffer);
  s->green_buffer = (unsigned char*) malloc(s->params.pixels_per_line*s->colour_buffer_lines*s->val[OPT_DEPTH]/8); 
  s->red_buffer = (unsigned char*) malloc(2*s->params.pixels_per_line*s->colour_buffer_lines*s->val[OPT_DEPTH]/8); 
  s->red_buf_count = s->green_buf_count = 0;
  interleave_colour = 0;
  start_col = SANE_UNFIX(s->val[OPT_TL_X]) * (1.0f * s->details->dpi/MM_PER_INCH) + 0x45; 
  end_col = start_col + (s->params.pixels_per_line*
			 (1.0f * s->details->dpi/s->val[OPT_RESOLUTION]));
 /* full steps to start of scan */
  start_line = s->details->y_offset + 
    (SANE_UNFIX((s->val[OPT_TL_Y] * s->details->fspi / MM_PER_INCH)));
  /* full steps during scan and a bit*/
  end_line = start_line +
    + ((10+s->params.lines + s->colour_buffer_lines*2) * (s->details->fspi / s->val[OPT_RESOLUTION]));
  DBG(3,"Geom: %d %d %d %d\n",start_col,start_line,end_col,end_line);
  /* set s->bytes_left to bytes in image */
  RETURN_IF_ERROR(set_parameters(s,s->val[OPT_DEPTH],s->val[OPT_RESOLUTION],
				 start_line,start_col,end_line,end_col,s->val[OPT_PREVIEW]));
  s->bytes_left = (s->params.lines + (s->colour_buffer_lines*2)) * s->params.bytes_per_line;
  /* set s->bytes_done to 0 */
  s->bytes_done = 0;
  s->lines_done = 0;
  /* wait for lamp to warm up */
  lamp_time = difftime(time(NULL),s->timer);
  while (lamp_time<5) {
    DBG(3,"Waiting for lamp to warm up: %gs\n",lamp_time);
    sleep(1);
    lamp_time = difftime(time(NULL),s->timer);
  }
  /* turn off paper_sensor */
  RETURN_IF_ERROR(paper_sensor1_off(s));
  /* calibrate */
#if 0
  if (s->val[OPT_CALIBRATE]) {
#endif
    switch (s->hw->scanner_make) {
    case LM9832_EPSON1250:
            RETURN_IF_ERROR(phils_calibration_bit(s,s->params.pixels_per_line));
      break;
    case LM9832_HP2200:
      DBG(1,"Calibrate\n");
      RETURN_IF_ERROR(do_coarse_calibration (s, &coarse));
      RETURN_IF_ERROR(do_fine_calibration (s, &coarse));
      break;
    }
#if 0
  }
#endif
  /* load gamma table */
  if (s->val[OPT_DEPTH]!=16) RETURN_IF_ERROR(set_gamma_table_flat(s));
  /* set everything in motion */
  print_registers(s);
  RETURN_IF_ERROR(sanei_lm9832_write_byte(s,0x07,3)); /* start scanner */
  /*print_registers(s);*/
  s->scanning = 1;
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_read (SANE_Handle handle, SANE_Byte *buf,
	   SANE_Int maxlen, SANE_Int *len)
{
  lm9832_Scanner *s = handle;
  SANE_Status status;
  int mylen;
  SANE_Byte temp;
  int offset = 0; /* offset into buffer */
  int line_bytes_remaining;
  int line_bytes_requested;
  sanei_lm9832_read_byte(s,1,&temp);
  DBG(3,"%dK in scanner, Reading %d, %d bytes left, %d bytes per line\n",temp,maxlen,s->bytes_left,s->params.bytes_per_line);
  if (s->scanning != SANE_TRUE || maxlen == 0)
    return SANE_STATUS_INVAL;
  //FIXME: need to account for 1/150" pitch between sensors...
  // work out how many lines we'll fit into this buffer...
  if (s->bytes_left > 0) {
    while (((maxlen-offset)>0) && (s->bytes_left>0)) {
      /* read rest of line into buf */
      DBG(5,"Reading6 %d %d\n",temp,offset);
      
      mylen = 0;
      line_bytes_remaining = 2 + s->params.bytes_per_line - s->bytes_done;
      line_bytes_requested = MIN(line_bytes_remaining,maxlen-offset);
      status = read_available_data(s,
				   buf + offset,
				   &mylen,
				   line_bytes_requested);
      DBG(5,"line: %x %x %x %x %x %x %x \n",
	  buf[offset],
	  buf[offset+1],
	  buf[offset+2],
	  buf[offset+3],
	  buf[offset+4],
	  buf[offset+5],
	  buf[offset+6]
	  );
      s->bytes_done += mylen;
      if (mylen == line_bytes_remaining){
	mylen -= 2;
	s->bytes_done = 0;
	s->lines_done++; /* increment line counter*/
	DBG(5,"5");
      }
      if (mylen == (1 + s->params.bytes_per_line - s->bytes_done)){
	status = sanei_lm9832_read(s,0,&buf[mylen+offset],1,0);
	s->bytes_done = 0;
	mylen -= 1;
	
      }
      if ((s->val[OPT_DEPTH]==16)&&(little_endian)) RETURN_IF_ERROR(make_little_endian(buf+offset,mylen));
      if (s->colour_buffer_lines) interleave(s,buf+offset,mylen); /* use buffered red and green data */
      offset += mylen; /* move offset into buf... */
      s->bytes_left -= mylen;
      if (s->lines_done < s->colour_buffer_lines*2) offset = 0; /* skip early data (just for buffering)*/
      if ((mylen < line_bytes_requested) && !s->blocking) break; /* skip out if little data returned... */
      if ((mylen == 0) && (offset>0)) break; /* skip for blocking areas*/
      if (status != SANE_STATUS_GOOD) break;
    }
  } else {
    status = SANE_STATUS_EOF; /* report end of file */
    offset = 0;
    do_cancel(s);
  }
  if (s->bytes_left <= 0) {
    DBG(3,"Ending\n");
  }
  *len = offset;
  return status;
}

void sane_cancel (SANE_Handle h) {
  lm9832_Scanner *s = h;
  do_cancel(s);
}

SANE_Status
sane_set_io_mode (SANE_Handle handle, SANE_Bool non_blocking)
{
  lm9832_Scanner *s = handle;
  DBG(3,"Non-blocking... %d\n",non_blocking);
  if (non_blocking) {
    return SANE_STATUS_UNSUPPORTED;
  } else {
    s->blocking = !non_blocking;
    return SANE_STATUS_GOOD;
  }
}

SANE_Status
sane_get_select_fd (SANE_Handle handle, SANE_Int * fd)
{
  return SANE_STATUS_UNSUPPORTED;
}
